"""
package: dgcv - Differential Geometry with Complex Variables
module: backends/_engine.py

Description: manages dgcv's interfacing with available CAS libraries.

Author (of this module): David Sykes (https://realandimaginary.com/dgcv/)

License:
    MIT License
"""

# -----------------------------------------------------------------------------
# imports and broadcasting
# -----------------------------------------------------------------------------
import importlib
import importlib.util

from .._config import dgcv_warning, get_dgcv_settings_registry

__all__ = [
    "is_sage_available",
    "is_sympy_available",
    "sage_module_if_available",
    "sympy_module_if_available",
    "available_engine_kinds",
    "engine_module",
    "engine_kind",
    "invalidate_engine_cache",
]


# -----------------------------------------------------------------------------
# utilities
# -----------------------------------------------------------------------------
_engine_kind = None
_engine_module = None
_sympy_module = None
_sage_module = None
_sage_available = None


def is_sage_available():
    global _sage_available
    if _sage_available is not None:
        return _sage_available
    try:
        spec = importlib.util.find_spec("sage.all")
        _sage_available = spec is not None
    except (ImportError, ModuleNotFoundError):
        _sage_available = False
    return _sage_available


def _get_sage_module():
    global _sage_module
    if _sage_module is not None:
        return _sage_module
    if not is_sage_available():
        raise RuntimeError("Sage is not available in the current environment.")
    _sage_module = importlib.import_module("sage.all")
    return _sage_module


def sage_module_if_available():
    if not is_sage_available():
        return None
    try:
        return _get_sage_module()
    except Exception:
        return None


def is_sympy_available():
    try:
        return importlib.util.find_spec("sympy") is not None
    except (ImportError, ModuleNotFoundError):
        return False


def _get_sympy_module():
    global _sympy_module
    if _sympy_module is not None:
        return _sympy_module
    if not is_sympy_available():
        raise RuntimeError("SymPy is not available in the current environment.")
    _sympy_module = importlib.import_module("sympy")
    return _sympy_module


def sympy_module_if_available():
    if not is_sympy_available():
        return None
    try:
        return _get_sympy_module()
    except Exception:
        return None


def available_engine_kinds():
    out = []
    if is_sage_available():
        out.append("sage")
    if is_sympy_available():
        out.append("sympy")
    return tuple(out)


def invalidate_engine_cache():
    global \
        _engine_kind, \
        _engine_module, \
        _one_obj, \
        _zero_obj, \
        _fast_scalar_types, \
        _expr_types, \
        _expr_numeric_types, \
        _atomic_pred, \
        _I_obj, \
        _constant_scalar_types
    _engine_kind = None
    _engine_module = None
    _fast_scalar_types = None
    _expr_types = None
    _expr_numeric_types = None
    _atomic_pred = None
    _one_obj = None
    _zero_obj = None
    _I_obj = None
    _constant_scalar_types = None
    try:
        from ._cls_coercion import invalidate_cls_coercion_cache

        invalidate_cls_coercion_cache()
    except Exception:
        pass


def _resolve_engine_kind():
    settings = get_dgcv_settings_registry()
    requested = str(settings.get("default_symbolic_engine", "sympy")).lower()

    if requested in ("sagemath",):
        requested = "sage"

    if requested == "sage":
        if is_sage_available():
            return "sage"
        if is_sympy_available():
            dgcv_warning(
                "dgcv: default symbolic engine setting is 'sage' but Sage is not available; "
                "falling back to 'sympy'.",
                stacklevel=2,
            )
            return "sympy"
        dgcv_warning(
            "dgcv: no symbolic engine was found, and dgcv requires one (either Sage or Sympy)",
            stacklevel=2,
        )
        return None

    if requested == "sympy":
        if is_sympy_available():
            return "sympy"
        if is_sage_available():
            dgcv_warning(
                "dgcv: default symbolic engine setting is 'sympy' but SymPy is not available; "
                "falling back to 'sage'.",
                stacklevel=2,
            )
            return "sage"
        dgcv_warning(
            "dgcv: no symbolic engine was found, and dgcv requires one (either Sage or Sympy)",
            stacklevel=2,
        )
        return None

    if is_sympy_available():
        dgcv_warning(
            f"dgcv: unrecognized symbolic engine {requested!r}; falling back to 'sympy'.",
            stacklevel=2,
        )
        return "sympy"

    if is_sage_available():
        dgcv_warning(
            f"dgcv: unrecognized symbolic engine {requested!r}; falling back to 'sage'.",
            stacklevel=2,
        )
        return "sage"

    dgcv_warning(
        f"dgcv: unrecognized symbolic engine {requested!r}, "
        "and no supported symbolic engine is available.",
        stacklevel=2,
    )
    return None


def engine_kind():
    global _engine_kind
    if _engine_kind is None:
        _engine_kind = _resolve_engine_kind()
    return _engine_kind


def engine_module():
    global _engine_module
    kind = engine_kind()
    if kind is None:
        return None
    if _engine_module is not None:
        return _engine_module
    if kind == "sage":
        _engine_module = _get_sage_module()
    else:
        _engine_module = _get_sympy_module()
    return _engine_module



# dgcv/backends/_display.py

from __future__ import annotations

import threading

from .._safeguards import check_dgcv_category
from ._engine import (
    _get_sympy_module,
    engine_kind,
    is_sage_available,
    is_sympy_available,
)

_tls = threading.local()


def _enter_guard(obj) -> bool:
    stack = getattr(_tls, "stack", None)
    if stack is None:
        stack = set()
        _tls.stack = stack
    oid = id(obj)
    if oid in stack:
        return False
    stack.add(oid)
    return True


def _exit_guard(obj) -> None:
    stack = getattr(_tls, "stack", None)
    if stack is None:
        return
    stack.discard(id(obj))


def _strip_math_delims(s: str) -> str:
    t = s.strip()
    if t.startswith("$$") and t.endswith("$$") and len(t) >= 4:
        return t[2:-2].strip()
    if t.startswith("$") and t.endswith("$") and len(t) >= 2:
        return t[1:-1].strip()
    return t


def _is_sympy_obj(x) -> bool:
    if not is_sympy_available():
        return False
    try:
        sp = _get_sympy_module()
        return isinstance(x, sp.Basic)
    except Exception:
        return False


def latex(expr, **kwargs) -> str:
    if expr is None:
        return ""

    if not _enter_guard(expr):
        try:
            return str(expr)
        except Exception:
            return repr(expr)

    try:
        if check_dgcv_category(expr):
            return _strip_math_delims(expr._repr_latex_(raw=True))

        if _is_sympy_obj(expr):
            try:
                from sympy.printing.latex import LatexPrinter

                return LatexPrinter(settings=kwargs).doprint(expr)
            except Exception:
                try:
                    sp = _get_sympy_module()
                    return sp.latex(expr, **kwargs)
                except Exception:
                    try:
                        return str(expr)
                    except Exception:
                        return repr(expr)

        if engine_kind() == "sage" and is_sage_available():
            try:
                from sage.misc.latex import latex as sage_latex

                return str(sage_latex(expr))
            except Exception:
                pass

        f = getattr(expr, "_repr_latex_", None)
        if callable(f):
            try:
                s = f()
                if isinstance(s, str):
                    return _strip_math_delims(s)
            except Exception:
                pass

        try:
            return str(expr)
        except Exception:
            return repr(expr)

    finally:
        _exit_guard(expr)



# backends/_exact_arith.py

import math
from numbers import Integral, Rational

from ._caches import _get_fast_scalar_types
from ._engine import _get_sage_module, _get_sympy_module, engine_kind
from ._types_and_constants import one, rational


def exact_reciprocal(x):
    if isinstance(x, _get_fast_scalar_types()):
        return one() / x

    if isinstance(x, Integral) and not isinstance(x, bool):
        return rational(1, x)

    if isinstance(x, Rational):
        try:
            return rational(x.numerator, x.denominator)
        except Exception:
            pass

    return one() / x


def factorial(n):
    f = getattr(n, "factorial", None)
    if f is not None:
        return f()

    if isinstance(n, Integral) and not isinstance(n, bool):
        return math.factorial(n)

    kind = engine_kind()

    if kind == "sympy":
        return _get_sympy_module().factorial(n)

    if kind == "sage":
        sage = _get_sage_module()
        return sage.factorial(n)

    return math.factorial(int(n))



"""
package: dgcv - Differential Geometry with Complex Variables
module: vector_fields_and_differential_forms

Description: This module provides tools for creating, manipulating, and decomposing vector fields and
differential forms within the dgcv package. It includes methods for Cartan calculus like
the exterior derivative and Lie derivative. There are some complex structure specific
functions as well, e.g., assembling holomorphic vector fields from holomorphic coefficients.

Key Functions:

Object Assembly:
    - get_VF(): Retrieves coordinate vector fields for the given coordinate variables.
    - get_DF(): Retrieves coordinate differential forms for the given coordinate variables.
    - assembleFromHolVFC(): Assembles a holomorphic vector field from holomorphic coefficients.
    - assembleFromAntiholVFC(): Assembles an antiholomorphic vector field from antiholomorphic
    coefficients.
    - assembleFromCompVFC(): Assembles a complex vector field from both holomorphic and
    antiholomorphic coefficients.

Differential Operators:
    - exteriorDerivative(): Computes the exterior derivative of a differential k-form.
    - interiorProduct(): Computes the interior product of a vector field with a differential k-form.
    - LieDerivative(): Computes the Lie derivative of a differential form or vector field with respect to another vector field.

Decompositions and Basis:
    - decompose(): Decomposes a vector field or differential form as a linear combination of a given basis of vector fields or differential forms.
    - get_coframe(): Constructs a coframe dual to a given list of vector fields.
    - annihilator(): Computes the annihilator (namely sub-bundle in TM/T^*M) of a list of differential forms or vector fields.

Author (of this module): David Sykes (https://realandimaginary.com/dgcv/)

License:
    MIT License
"""

# -----------------------------------------------------------------------------
# imports and broadcasting
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import Any, Dict, List, Optional, Sequence, Tuple

from ._config import dgcv_warning, get_dgcv_settings_registry
from ._safeguards import (
    create_key,
    get_dgcv_category,
    query_dgcv_categories,
    retrieve_passkey,
)
from .backends._numeric_router import zeroish
from .backends._symbolic_router import _scalar_is_zero, get_free_symbols, simplify, subs
from .backends._types_and_constants import expr_numeric_types, symbol
from .combinatorics import chooseOp
from .conversions import allToReal
from .dgcv_core import (
    VF_bracket,
    differential_form_class,
    variableProcedure,
    vector_field_class,
    wedge,
)
from .solvers import solve_dgcv
from .vmf import vmf_lookup


# -----------------------------------------------------------------------------
# retrieval
# -----------------------------------------------------------------------------
def get_VF(*coordinates):
    out = []
    for coord in coordinates:
        info = vmf_lookup(coord, relatives=True, differential_system=True)

        if info.get("type") != "coordinate":
            continue

        ds = info.get("differential_system")
        if not isinstance(ds, dict):
            continue

        vf = ds.get("vf")
        if vf is None:
            continue

        out.append(vf)

    return out


def get_DF(*coordinates):
    out = []
    for coord in coordinates:
        info = vmf_lookup(coord, relatives=True, differential_system=True)

        if info.get("type") != "coordinate":
            continue

        ds = info.get("differential_system")
        if not isinstance(ds, dict):
            continue

        df = ds.get("df")
        if df is None:
            continue

        out.append(df)

    return out


def coordinate_vector_field(obj):
    vf_list = get_VF(obj)
    if len(vf_list) != 1:
        return
    return vf_list[0]


def coordinate_differential_form(obj):
    df_list = get_DF(obj)
    if len(df_list) != 1:
        return
    return df_list[0]


# -----------------------------------------------------------------------------
# complex vector fields
# -----------------------------------------------------------------------------
def _require_complex_coordinate_atom(atom: Any, *, context: str) -> dict:
    info = vmf_lookup(atom, path=True, relatives=True)
    if info.get("type") != "coordinate":
        raise TypeError(
            f"{context} requires variables registered in the VMF as coordinates. "
            "Suggestion: initialize complex coordinate systems with dgcv.createVariables(...)."
        )
    st = info.get("sub_type")
    if st == "standard":
        raise TypeError(
            f"{context} requires complex coordinates (holo/anti/real/imag), "
            "but received a standard coordinate. Suggestion: use dgcv.createVariables(...) "
            "to register a complex coordinate system in the VMF."
        )
    return info


def _pick_relative(atom: Any, *, want: str, context: str):
    info = _require_complex_coordinate_atom(atom, context=context)
    st = info.get("sub_type")

    rel = info.get("relatives") or {}
    if want == "holo":
        if st == "holo":
            out = atom
        else:
            out = rel.get("holo", None)
    elif want == "anti":
        if st == "anti":
            out = atom
        else:
            out = rel.get("anti", None)
    elif want == "real":
        if st == "real":
            out = atom
        else:
            out = rel.get("real", None)
    elif want == "imag":
        if st == "imag":
            out = atom
        else:
            out = rel.get("imag", None)
    else:
        raise ValueError("want must be one of: 'holo', 'anti', 'real', 'imag'")

    if out is None:
        raise TypeError(
            f"{context} requires the variable to belong to a complex coordinate system "
            "with the appropriate relatives available in the VMF."
        )

    return out, st


def _warn_if_not_holo_inputs(hol_vars: Sequence[Any], *, context: str):
    if get_dgcv_settings_registry().get("forgo_warnings", False):
        return

    bad = []
    for v in hol_vars:
        info = vmf_lookup(v, relatives=False)
        if info.get("type") != "coordinate":
            continue
        st = info.get("sub_type")
        if st not in (None, "holo"):
            bad.append((v, st))

    if bad:
        dgcv_warning(
            f"{context}: holomorphic variables (sub_type='holo') are recommended for hol_vars. "
            f"Received non-holo coordinate atoms: {bad}. "
            "Proceeding by using VMF relatives to interpret them."
        )


def assembleFromHolVFC(
    coeffs: Sequence[Any],
    hol_vars: Sequence[Any],
    *,
    _warn_on_nonholo: bool = True,
) -> "vector_field_class":
    if not isinstance(hol_vars, (list, tuple)):
        raise TypeError("hol_vars must be a list or tuple.")
    if len(coeffs) != len(hol_vars):
        raise ValueError("coeffs and hol_vars must have the same length.")
    if _warn_on_nonholo:
        _warn_if_not_holo_inputs(hol_vars, context="assembleFromHolVFC")

    picked = [
        (_pick_relative(z, want="holo", context="assembleFromHolVFC")[0])
        for z in hol_vars
    ]
    Dz_list = get_VF(*picked)

    out = None
    for c, Dz in zip(coeffs, Dz_list):
        if not c:
            continue
        term = c * Dz
        out = term if out is None else (out + term)

    if out is None:
        out = vector_field_class(coeff_dict={tuple(): 0}, data_shape="all")

    return out


def assembleFromAntiholVFC(
    coeffs: Sequence[Any],
    hol_vars: Sequence[Any],
    *,
    _warn_on_nonholo: bool = True,
) -> "vector_field_class":
    if not isinstance(hol_vars, (list, tuple)):
        raise TypeError("hol_vars must be a list or tuple.")
    if len(coeffs) != len(hol_vars):
        raise ValueError("coeffs and hol_vars must have the same length.")
    if _warn_on_nonholo:
        _warn_if_not_holo_inputs(hol_vars, context="assembleFromAntiholVFC")

    picked = [
        (_pick_relative(z, want="anti", context="assembleFromAntiholVFC")[0])
        for z in hol_vars
    ]
    Dzb_list = get_VF(*picked)

    out = None
    for c, Dzb in zip(coeffs, Dzb_list):
        if not c:
            continue
        term = c * Dzb
        out = term if out is None else (out + term)

    if out is None:
        out = vector_field_class(coeff_dict={tuple(): 0}, data_shape="all")

    return out


def assembleFromCompVFC(
    holomorphic_coeffs: Sequence[Any],
    antiholomorphic_coeffs: Sequence[Any],
    hol_vars: Sequence[Any],
    *,
    _warn_on_nonholo: bool = True,
) -> "vector_field_class":
    if not isinstance(hol_vars, (list, tuple)):
        raise TypeError("hol_vars must be a list or tuple.")
    if len(holomorphic_coeffs) != len(hol_vars) or len(antiholomorphic_coeffs) != len(
        hol_vars
    ):
        raise ValueError("Coefficient lengths must match hol_vars length.")

    vf_h = assembleFromHolVFC(
        holomorphic_coeffs, hol_vars, _warn_on_nonholo=_warn_on_nonholo
    )
    vf_a = assembleFromAntiholVFC(
        antiholomorphic_coeffs, hol_vars, _warn_on_nonholo=_warn_on_nonholo
    )
    return vf_h + vf_a


# -----------------------------------------------------------------------------
# differential forms
# -----------------------------------------------------------------------------
def makeZeroForm(
    expr: Any,
    varSpace: Sequence[Any] | None = None,
    *,
    variable_spaces: dict | None = None,
    data_shape: str = "all",
    dgcvType: str = "standard",
    _simplifyKW=None,
    parameters=set(),
) -> "differential_form_class":
    if variable_spaces is None:
        rv = get_free_symbols(expr)
        if rv:
            params = getattr(expr, "parameters", None)
            if params:
                rv = set(rv) - set(params)
        else:
            rv = set()

        variable_spaces = _prep_symb_set_for_ext_der(
            rv, use_for_zero_form=True, full_dict=True
        )

    return differential_form_class(
        coeff_dict={tuple(): expr},
        valence=tuple(),
        data_shape=data_shape,
        dgcvType=dgcvType,
        _simplifyKW=_simplifyKW,
        variable_spaces=variable_spaces,
        parameters=parameters,
    )


def _prep_symb_set_for_ext_der(symbols, use_for_zero_form=False, full_dict=False):
    registry: Dict[Any, Any] = {}

    for atom in symbols:
        atom_data = vmf_lookup(
            atom,
            path=True,
            flattened_relatives=True,
            differential_system=True,
        )
        if atom_data.get("type") != "coordinate":
            continue

        path = atom_data.get("path")
        if not (isinstance(path, tuple) and len(path) >= 2):
            continue

        syslbl = path[1]

        if full_dict or use_for_zero_form:
            if syslbl in registry:
                continue

            sys_key = atom_data.get("differential_system") or syslbl
            sys_data = vmf_lookup(sys_key, flattened_relatives=True)

            flattened = sys_data.get("flattened_relatives")
            if isinstance(flattened, tuple) and flattened:
                registry[syslbl] = flattened
            continue

        sysreg = registry.setdefault(syslbl, {})

        if atom in sysreg:
            continue

        st = atom_data.get("sub_type")
        rel = atom_data.get("relatives") or {}

        if st in {"holo", "anti"}:
            sysreg[atom] = (rel.get("holo"), rel.get("anti"))
        elif st in {"real", "imag"}:
            sysreg[atom] = (rel.get("holo"), rel.get("anti"))
        else:
            ds = atom_data.get("differential_system")
            if ds is not None:
                sysreg[atom] = (atom,)

    return registry


def exteriorDerivative(form_or_scalar: Any, **kwargs) -> "differential_form_class":
    obj = form_or_scalar

    if get_dgcv_category(obj) != "tensor_field":
        if isinstance(obj, expr_numeric_types()):
            obj = makeZeroForm(obj)
        else:
            raise TypeError(
                "exteriorDerivative expects a differential form or scalar expression."
            )

    if not query_dgcv_categories(obj, {"differential_form"}):
        raise TypeError(
            "exteriorDerivative expects a differential form or scalar expression."
        )

    form = obj

    rv = get_free_symbols(form)
    if rv:
        params = getattr(form, "parameters", None)
        if params:
            rv = set(rv) - set(params)
    else:
        rv = set()

    relevant_vars = _prep_symb_set_for_ext_der(rv)

    accumulation = 0
    for system in relevant_vars.values():
        for atom in system:
            ds = vmf_lookup(atom, differential_system=True).get("differential_system")
            if ds is None:
                continue
            vf = ds.get("vf")
            df = ds.get("df")
            if vf is None or df is None:
                continue
            accumulation += df * (form.apply(vf))

    return accumulation


def interiorProduct(
    vector_field: "vector_field_class",
    differential_form: "differential_form_class",
) -> "differential_form_class":
    if not query_dgcv_categories(vector_field, {"vector_field"}):
        raise TypeError("interiorProduct expects a vector_field_class instance.")
    if not query_dgcv_categories(differential_form, {"differential_form"}):
        raise TypeError("interiorProduct expects a differential_form_class instance.")

    return differential_form(vector_field)


def LieDerivative(
    vector_field: "vector_field_class",
    obj: Any,
    **kwargs,
):
    if not query_dgcv_categories(vector_field, {"vector_field"}):
        raise TypeError(
            "LieDerivative expects the first argument to be a vector field."
        )

    if isinstance(obj, expr_numeric_types()):
        return vector_field(obj)

    if get_dgcv_category(obj) != "tensor_field":
        raise TypeError("LieDerivative expects a tensor field or scalar expression.")

    if query_dgcv_categories(obj, {"differential_form"}):
        return exteriorDerivative(interiorProduct(vector_field, obj)) + interiorProduct(
            vector_field, exteriorDerivative(obj)
        )

    if query_dgcv_categories(obj, {"vector_field"}):
        return VF_bracket(vector_field, obj)

    raise TypeError(
        "LieDerivative currently supports differential forms and vector fields only."
    )


# -----------------------------------------------------------------------------
# vector field and differential forms operations with linear solve computations
# -----------------------------------------------------------------------------
def _local_unknowns(n: int, lbl: str = "__tem_label__") -> List[Any]:
    return [symbol(f"{lbl}{i}") for i in range(n)]


def _free_vars_from_solution(
    sol: Dict[Any, Any], vars_list: Sequence[Any]
) -> List[Any]:
    out: List[Any] = []
    for v in vars_list:
        try:
            if sol.get(v, None) == v:
                out.append(v)
        except Exception:
            pass
    return out


def _sub_free_to_zero(sol: Dict[Any, Any], free: Sequence[Any]) -> Dict[Any, Any]:
    if not free:
        return sol
    out = dict(sol)
    for v in free:
        out[v] = 0
    return out


def _make_parameters(
    n: int,
    *,
    register_parameters: bool,
    parameters_label: Optional[str],
) -> Tuple[Any, ...]:
    if n <= 0:
        return tuple()

    if register_parameters:
        prefix = parameters_label if isinstance(parameters_label, str) else "param"
        params = variableProcedure(prefix, n, return_created_object=True)[0]
        return tuple(params)

    prefix = parameters_label if isinstance(parameters_label, str) else "p"
    return tuple(symbol(f"{prefix}{i}") for i in range(n))


def _require_subcategory(objs: Sequence[Any], cats: set[str], who: str) -> None:
    for o in objs:
        if not query_dgcv_categories(o, cats):
            raise TypeError(
                f"`{who}` expects dgcv objects in categories {sorted(cats)}."
            )


def _key_universe(tfs):
    K = set()
    for tf in tfs:
        if get_dgcv_category(tf) == "tensor_field":
            for k, v in tf.coeff_dict.items():
                if v != 0:
                    K.add(k)
    return tuple(sorted(K))


def _as_coeff_vector_form(tf, K, syslbl="__dgcv_par__"):
    cd = {}
    varspacPlaceholder = None
    for j, k in enumerate(K):
        v = tf.coeff_dict.get(k, 0)
        if v != 0:
            cd[(j, 0, syslbl)] = v
    if not cd:
        cd = {tuple(): 0}
    return differential_form_class(
        coeff_dict=cd, data_shape="skew", variable_spaces={syslbl: varspacPlaceholder}
    )


def _extract_basis_by_wedge_vectorized(
    objs, *, use_numeric_methods: bool = False, dimension_hint=None
):
    if not objs:
        return []
    if dimension_hint is None:
        dimension_hint = len(objs)

    use_numeric = use_numeric_methods or bool(
        get_dgcv_settings_registry().get("use_numeric_methods", False)
    )

    def _is_zero_scalar(x):
        if use_numeric:
            return zeroish(x)
        return _scalar_is_zero(x)

    if all(get_dgcv_category(o) == "array" for o in objs):
        K = set()
        for o in objs:
            K.update(getattr(o, "_data", {}).keys())
        if not K:
            return []

        K = tuple(sorted(K))

        vecs = []
        for o in objs:
            d = getattr(o, "_data", {})
            v = {}
            for k in K:
                val = d.get(k, 0)
                if val is None or _is_zero_scalar(val):
                    continue
                v[k] = val
            vecs.append(v)

        def _kform_is_zero(F):
            return not F or all(_is_zero_scalar(c) for c in F.values())

        def _wedge_kform_vector(F, v):
            out = {}
            if not F or not v:
                return out
            for idx, c in F.items():
                I_set = set(idx)
                for j, vj in v.items():
                    if j in I_set:
                        continue
                    pos = 0
                    for t in idx:
                        if t < j:
                            pos += 1
                        else:
                            break
                    sign = -1 if ((len(idx) - pos) % 2) else 1
                    J = idx[:pos] + (j,) + idx[pos:]
                    coeff = sign * c * vj
                    if _is_zero_scalar(coeff):
                        continue
                    out[J] = out.get(J, 0) + coeff
            if not use_numeric:
                out = {k: simplify(v) for k, v in out.items() if not _is_zero_scalar(v)}
            else:
                out = {k: v for k, v in out.items() if not _is_zero_scalar(v)}
            return out

        obstruction = None
        out = []
        for o, v in zip(objs, vecs):
            if not v:
                continue

            if obstruction is None:
                obstruction = {(k,): val for k, val in v.items()}
                if not use_numeric:
                    obstruction = {
                        idx: simplify(c)
                        for idx, c in obstruction.items()
                        if not _is_zero_scalar(c)
                    }
                out.append(o)
                continue

            w = _wedge_kform_vector(obstruction, v)
            if _kform_is_zero(w):
                continue

            obstruction = w
            out.append(o)

        return out

    K = _key_universe(objs)
    if not K:
        return []

    try:
        vec_forms = [_as_coeff_vector_form(o, K) for o in objs]
    except Exception as exc:
        raise TypeError(
            "Could not compute a linear independence test for these objects because "
            "they do not support the required linear combination behavior."
        ) from exc

    obstruction = None
    out = []
    for o, v in zip(objs, vec_forms):
        if len(out) == dimension_hint:
            break
        if use_numeric:
            if zeroish(v):
                continue
        else:
            if _scalar_is_zero(v):
                continue

        if obstruction is None:
            obstruction = v
            out.append(o)
            continue

        w = wedge(obstruction, v)
        if not use_numeric:
            w = simplify(w)
            if _scalar_is_zero(w):
                continue
        else:
            if zeroish(w):
                continue

        obstruction = w
        out.append(o)

    return out


def _coordinate_basis_from_vmf(
    coordinate_space: Sequence[Any],
    *,
    want: str,  # "df" or "vf"
) -> List[Any]:
    out: List[Any] = []
    for atom in coordinate_space:
        info = vmf_lookup(atom, differential_system=True)
        ds = info.get("differential_system")
        if not isinstance(ds, dict):
            raise TypeError(
                "`coordinate_space` must consist of VMF-registered coordinates with a differential system."
            )
        obj = ds.get(want)
        if obj is None:
            raise TypeError(
                "`coordinate_space` must consist of VMF-registered coordinates with a differential system."
            )
        out.append(obj)
    return out


def _infer_coordinate_space_from_objs(objs: Sequence[Any]) -> List[Any]:
    syms: set[Any] = set()
    for o in objs:
        fs = get_free_symbols(o)
        if fs:
            params = getattr(o, "parameters", None)
            if params:
                fs = set(fs) - set(params)
            syms |= set(fs)

    atoms: List[Any] = []
    for a in syms:
        info = vmf_lookup(a, differential_system=True)
        ds = info.get("differential_system")
        if isinstance(ds, dict) and (
            ds.get("vf") is not None or ds.get("df") is not None
        ):
            atoms.append(a)

    atoms.sort(key=lambda x: str(x))
    return atoms


def decompose(
    obj,
    basis,
    return_parameters: bool = False,
    new_parameters_label: str | None = None,
    only_check_decomposability: bool = False,
    _pass_error_report=None,
    _hand_off=None,
    *,
    assume_basis: bool = False,
    register_parameters: bool = False,
):
    """
    Decomposes a vector field or differential form as a linear combination of a given `basis` list.

    This function attempts to express the input `obj` (a VFClass or DFClass object) as a linear combination
    of the elements in the provided `basis` list. The `basis` list does not need to be linearly independent,
    and if the decomposition is not unique, the function will parameterize the solution space. Any parameters
    needed are automatically initialized and registered in dgcv's variable management framework (VMF).

    The function carefully handles variable types based on the `dgcvType` attribute of the objects. For objects
    with `dgcvType='complex'`, it dynamically selects whether to perform real coordinate computations or complex
    coordinate computations, depending on the input data. If a canonical variable formatting decision cannot be
    made naturally from the input, the function will return warnings with explanations.

    Parameters
    ----------
    obj : VFClass or DFClass
        The vector field or differential form to decompose.
    basis : list of VFClass or DFClass
        A list of objects (vector fields or differential forms) to decompose `obj` with respect to.
        The class of objects in the `basis` list must match the class of `obj`.
    return_parameters : bool, optional
        If True, the function will return the parameterized solution when the decomposition is not unique
        (default is False). Parameters are initialized with labels registered within the VMF.
    new_parameters_label : str or None, optional
        If `return_parameters` is True and the decomposition is not unique, this label will be used
        to name the new parameter variables. If None, obscure labels will be generated automatically (default is None).
    _pass_error_report : optional
        Internal use parameter for handling error reports in certain edge cases (default is None).

    Returns
    -------
    list
        The coefficients of the linear combination that expresses `obj` in terms of the `basis` list.
        If the decomposition is parameterized, the returned list contains the parameterized solution.

    Raises
    ------
    TypeError
        If the class of `obj` does not match the class of elements in the `basis` list (i.e., both must
        be either VFClass or DFClass), or if objects in the `basis` list have inconsistent `dgcvType` attributes.

    Warnings
    --------
    - A warning is issued if `obj` is not in the span of the provided `basis` list.
    - If the `basis` list is not linearly independent, the decomposition is not unique, and a parameterized
    solution will be returned. The initialized parameters are registered as 'standard variables' in the VMF.

    Remarks
    -------
    - The function dynamically handles objects based on their `dgcvType` attribute. For `dgcvType='complex'`,
    it distinguishes between real and complex coordinate computations, converting the input as needed
    to ensure consistency in formatting. If this decision cannot be determined from the input data,
    the function issues warnings explaining the necessary canonical formatting.

    Example
    -------
    # Decompose a vector field 'vf' as a linear combination of two basis vector fields 'vf1' and 'vf2'
    coeffs, basis_used = decompose(vf, [vf1, vf2])

    # Decompose a differential form 'df' with a parameterized solution due to non-uniqueness
    coeffs, basis_used = decompose(df, [df1, df2, df3], return_parameters=True, new_parameters_label='p')
    """
    basis = list(basis)
    original_length = len(basis)

    if not return_parameters and not assume_basis:
        basis = _extract_basis_by_wedge_vectorized(basis)

    n = len(basis)
    if n == 0:
        if only_check_decomposability is True:
            ob = getattr(obj, "__dgcv_zero_obstr__", None)
            if ob is None:
                raise TypeError(
                    f"decompose does not operate on objects of type {type(obj).__name__}"
                )
            eqns, _ = ob
            eqns = list(eqns or [])
            return bool(eqns) and all(_scalar_is_zero(e) for e in eqns)
        return ([], basis)

    label = create_key(prefix="var")
    vars = [symbol(f"{label}{i}") for i in range(n)]

    try:
        system = obj
        for v, e in zip(vars, basis):
            system = system - v * e
    except Exception as exc:
        msg = (
            "`decompose` could not form the residual `obj - sum(var_i * basis_i)` "
            f"for obj type {type(obj).__name__} and basis element types "
            f"{[type(b).__name__ for b in basis]}. "
            "This usually means the objects do not support compatible addition/subtraction "
            "and scalar multiplication."
        )
        if _pass_error_report == retrieve_passkey():
            return msg + f" Original error: {exc!r}"
        raise TypeError(msg) from exc

    ob = getattr(system, "__dgcv_zero_obstr__", None)
    if ob is None:
        raise TypeError(f"{type(system).__name__} lacks __dgcv_zero_obstr__")
    eqns, _ = ob
    eqns = list(eqns or [])

    solutions = solve_dgcv(eqns, vars, method="linsolve")

    if not solutions:
        if only_check_decomposability is True:
            return False
        if _pass_error_report == retrieve_passkey():
            return (
                f"`decompose` found no solution for equations {eqns} in variables {vars} "
                f"against a spanning family of length {original_length}."
            )
        return ([], basis)

    if only_check_decomposability is True:
        return True

    sol0 = solutions[0]
    decomp_coeffs = [sol0.get(v, v) for v in vars]

    free = set()
    for expr in decomp_coeffs:
        free |= set(get_free_symbols(expr) or ())

    if return_parameters and free:
        free = tuple(free)
        if register_parameters:
            params = variableProcedure(
                new_parameters_label
                if isinstance(new_parameters_label, str)
                else "param",
                len(free),
                return_created_object=True,
            )[0]
            params = tuple(params)
        else:
            plabel = (
                new_parameters_label
                if isinstance(new_parameters_label, str) and new_parameters_label
                else create_key(prefix="param")
            )
            params = tuple(symbol(f"{plabel}{i}") for i in range(len(free)))

        subs_dict = {v: p for v, p in zip(free, params)}
        decomp_coeffs = [subs(c, subs_dict) for c in decomp_coeffs]

    return (decomp_coeffs, basis)


def get_coframe(
    VFList: Sequence[Any],
    *,
    coordinate_space: Sequence[Any] | None = None,
    return_parameters: bool = False,
    register_parameters: bool = False,
    parameters_label: str | None = None,
):
    VFList = list(VFList)
    if not VFList:
        return []

    _require_subcategory(VFList, {"vector_field"}, "get_coframe")

    if coordinate_space is None:
        coordinate_space = _infer_coordinate_space_from_objs(VFList)
    else:
        if not isinstance(coordinate_space, (list, tuple, set)):
            raise TypeError(
                "`get_coframe` coordinate_space must be a list/tuple/set if provided."
            )
        coordinate_space = list(coordinate_space)
        for a in coordinate_space:
            ds = vmf_lookup(a, differential_system=True).get("differential_system")
            if ds is None:
                raise TypeError(
                    "`get_coframe` requires coordinates to be registered in the dgcv VMF with differential objects. "
                    "Suggestion: initialize coordinates with `createVariables(..., withVF=True)` (or `createVariables(..., complex=True)` as appropriate)."
                )

    dfBasis = _coordinate_basis_from_vmf(coordinate_space, want="df")
    n = len(dfBasis)
    m = len(VFList)

    if n == 0:
        raise TypeError("`get_coframe` could not obtain a DF basis from VMF.")

    def _as_scalar_eqns(x):
        if get_dgcv_category(x) == "tensor_field":
            return list(x.coeff_dict.values())
        return [x]

    eval_entries: List[List[Any]] = []
    for k in range(m):
        row = []
        vf = VFList[k]
        for df in dfBasis:
            val = df(vf)
            scalars = _as_scalar_eqns(val)
            if len(scalars) != 1:
                row.append(sum(scalars))
            else:
                row.append(scalars[0])
        eval_entries.append(row)

    unknowns = _local_unknowns(m * n)

    eqns: List[Any] = []
    for j in range(m):
        row_vars = unknowns[j * n : (j + 1) * n]
        for k in range(m):
            target = 1 if j == k else 0
            s = 0
            Ek = eval_entries[k]
            for u, e in zip(row_vars, Ek):
                s = s + u * e
            eqns.append(s - target)

    sols = solve_dgcv(eqns, unknowns, method="linsolve")
    if not sols:
        raise RuntimeError(
            "`get_coframe` could not solve for a coframe (system unsatisfiable)."
        )

    sol = sols[0]

    unknowns_set = set(unknowns)
    free_vars_set = set()
    for u in unknowns:
        free_vars_set |= set(get_free_symbols(sol.get(u, u))) & unknowns_set
    free = [u for u in unknowns if u in free_vars_set]

    if return_parameters and free:
        params = _make_parameters(
            len(free),
            register_parameters=register_parameters,
            parameters_label=parameters_label,
        )
        sub = dict(zip(free, params))
        sol_use = dict(sol)
        for v in free:
            sol_use[v] = v

        def _coeff(u):
            c = sol_use.get(u, u)
            return subs(c, sub)
    else:
        sol_use = dict(sol)
        for v in free:
            sol_use[v] = 0

        def _coeff(u):
            return sol_use.get(u, u)

    out: List[Any] = []
    for j in range(m):
        row_vars = unknowns[j * n : (j + 1) * n]
        omega_j = 0
        for u, df in zip(row_vars, dfBasis):
            omega_j = omega_j + _coeff(u) * df
        out.append(omega_j)

    return out


def annihilator(
    objList: Sequence[Any],
    coordinate_space: Sequence[Any] | None = None,
    control_distribution: Sequence[Any] | None = None,
    polynomial_bases: bool = False,
    _pass_error_report=None,
    allow_div_by_zero: bool = False,
    *,
    return_parameters: bool = False,
    register_parameters: bool = False,
    parameters_label: str | None = None,
    coherent_coordinates_checked: bool = False,
):
    """
    Finds annihilators for a given list of vector fields or differential forms.

    This function computes objects that "annihilate" the provided list of vector fields or differential forms.
    An annihilator is either the span of differential forms that evaluate to zero on each vector field in the list,
    or vector fields whose interior product annihilates each differential form in the list. `annihilator` dynamically
    handles both real and holomorphic coordinate systems and can convert between them as needed. Additionally,
    solutions can be constrained to a given control distribution by using the `control_distribution` keyword.

    Parameters
    ----------
    objList : list of VFClass or DFClass
        A list of vector fields or differential forms for which the annihilator will be computed. All objects
        in the list must be of the same class (either all vector fields or all differential forms) and have
        consistent `dgcvType` attributes (i.e., 'standard' or 'complex').
    coordinate_Space : list, tuple, or set
        A collection of variables that define the coordinate system in which the annihilator is to be computed.
    polynomial_bases: bool (optional)
        Attempt to scale computed basis elements so that they have polynomial coefficients before returning the list. Can still produce non-nonpolynomial expressions when factoring is not possible.
    allow_div_by_zero : bool, optional
        If True, allows the annihilator to be returned without scaling to avoid division by zero (default is False).
        Scaling to avoid division has more computational overhead but typically simplifies output.
    _pass_error_report : optional
        Internal use parameter for handling error reports in certain edge cases (default is None).

    Returns
    -------
    list
        A list of differential forms (if vector fields were provided) or vector fields (if differential forms were
        provided) that annihilate the objects in `objList`.

    Raises
    ------
    TypeError
        If the objects in `objList` are not all of the same type (i.e., all vector fields or all differential forms),
        or if the `coordinate_Space` is not a valid list, tuple, or set.

    Warnings
    --------
    - If the objects in `objList` are defined with inconsistent coordinate system types (real vs. holomorphic), the
    function converts them to a consistent coordinate system and issues a warning.

    Example
    -------
    >>> from dgcv import createVariables, annihilator, exteriorDerivative, complex_struct_op, Del, DelBar, allToReal
    >>> createVariables('z', 'x', 'y', 4, initialIndex=0)
    >>> rho = (x1*x2 + x1**2*x3 - x0)  # A defining equation for a real hypersurface M in C^4
    >>> d_rho = exteriorDerivative(rho)  # Its differential will annihilate TM
    >>> print(d_rho)

    (2*x1*x3 + x2)*d_x1 + x1*d_x2 + x1**2*d_x3 - 1*d_x0

    >>> dfList = [d_rho]
    >>> TMbasis = annihilator(dfList, x+y)  # Use annihilator to compute the tangent bundle TM
    >>> TMbasis

    [(16*x1*x3 + 8*x2)*D_x0 + 8*D_x1, 8*x1*D_x0 + 8*D_x2, 8*x1**2*D_x0 + 8*D_x3, 8*D_y0, 8*D_y1, 8*D_y2, 8*D_y3]

    >>> J_of_TMbasis = [complex_struct_op(vf) for vf in TMbasis]  # Get the image of TM under the complex structure operator.
    >>> J_of_TMbasis

    [(16*x1*x3 + 8*x2)*D_y0 + 8*D_y1, 8*x1*D_y0 + 8*D_y2, 8*x1**2*D_y0 + 8*D_y3, -8*D_x0, -8*D_x1, -8*D_x2, -8*D_x3]

    >>> CR_distribution = annihilator(annihilator(J_of_TMbasis, x+y) + annihilator(TMbasis, x+y), x+y)
    >>> CR_distribution  # Use annihilator to get the CR distribution, which is the intersection of TM with CTM

    [(16*x1*x3 + 8*x2)*D_x0 + 8*D_x1, 8*x1*D_x0 + 8*D_x2, 8*x1**2*D_x0 + 8*D_x3,
    (16*x1*x3 + 8*x2)*D_y0 + 8*D_y1, 8*x1*D_y0 + 8*D_y2, 8*x1**2*D_y0 + 8*D_y3]

    >>> LeviForm = allToReal(Del(DelBar(rho)))  # Apply Dolbeault operators to represent the Levi form
    >>> print(LeviForm._repr_latex_())

    # Output: <LeviForm in LaTeX formatted plain text>

    >>> Levi_kernel = annihilator([LeviForm], x+y, control_distribution=CR_distribution)
    >>> Levi_kernel  # annihilator reveals that the Levi form has a real 2-d. kernel

    [-64*x1**2*D_x0 - 128*x1*D_x2 + 64*D_x3, -64*x1**2*D_y0 - 128*x1*D_y2 + 64*D_y3]

    >>> not_the_Levi_kernel = annihilator([LeviForm], x+y)
    >>> not_the_Levi_kernel  # Without constraining annihilator to the CR distribution, it finds a kernel that is too large.

    [8*D_x0, -16*x1*D_x2 + 8*D_x3, 8*D_y0, -16*x1*D_y2 + 8*D_y3]
    """
    objList = list(objList)
    if not objList:
        return []

    if not all(get_dgcv_category(o) == "tensor_field" for o in objList):
        raise TypeError("`annihilator` expects tensor_field objects.")

    is_vf_case = all(query_dgcv_categories(o, {"vector_field"}) for o in objList)
    control_needed, expensive_path_needed = control_distribution is not None, False
    if is_vf_case is True:
        is_df_case = False
    else:
        is_df_case = True
        for o in objList:
            if not query_dgcv_categories(o, {"differential_form"}):
                is_df_case = False
                break
            md, mmd = o.min_degree, o.max_degree
            if control_needed and mmd > 1:
                expensive_path_needed = True
                if md != mmd:
                    raise TypeError(
                        "`annihilator` does not support restriction to control distributions when the objects to annihilate are mixed deree differential forms."
                    )
    if not is_vf_case and not is_df_case:
        raise TypeError(
            "`annihilator` expects a list of all vector fields or all differential forms."
        )
    if control_needed:
        if is_vf_case:
            if get_dgcv_category(control_distribution) == "distribution":
                control_distribution = control_distribution.df_basis
            else:
                control_distribution = list(control_distribution)
                if not all(
                    query_dgcv_categories(o, {"differential_form"})
                    for o in control_distribution
                ):
                    raise TypeError(
                        "`annihilator` control_distribution must be differential forms if given objects to annihilate are VF."
                    )
            basis_elems = control_distribution
        else:
            if get_dgcv_category(control_distribution) == "distribution":
                control_distribution = control_distribution.vf_basis
            else:
                control_distribution = list(control_distribution)
                if not all(
                    query_dgcv_categories(o, {"vector_field"})
                    for o in control_distribution
                ):
                    raise TypeError(
                        "`annihilator` control_distribution must be vector fields if the given objects to annihilater are DF."
                    )
            basis_elems = control_distribution

    elif coordinate_space is not None:
        if not isinstance(coordinate_space, (list, tuple, set)):
            raise TypeError(
                "`annihilator` coordinate_space must be a list/tuple/set if provided."
            )

        coords = list(coordinate_space)
        for a in coords:
            ds = vmf_lookup(a, differential_system=True).get("differential_system")
            if ds is None:
                raise TypeError(
                    "`annihilator` requires manually provided coordinates to be registered in the dgcv VMF with differential objects. "
                    "Suggestion: initialize coordinates with `createVariables(..., withVF=True)` (or `createVariables(..., complex=True)` as appropriate)."
                )

        basis_elems = []
        for a in coords:
            ds = vmf_lookup(a, differential_system=True).get("differential_system")
            elem = ds.get("df") if is_vf_case else ds.get("vf")
            if elem is None:
                raise TypeError(
                    "`annihilator` requires coordinates to be registered in the dgcv VMF with differential objects. "
                    "Suggestion: initialize coordinates with `createVariables(..., withVF=True)` (or `createVariables(..., complex=True)` as appropriate)."
                )
            basis_elems.append(elem)

    else:
        atoms = set()
        for o in objList:
            vs = getattr(o, "_variable_spaces", None)
            if isinstance(vs, dict):
                for tup in vs.values():
                    atoms.update(tup)

        basis_elems = []
        for a in sorted(atoms, key=lambda x: str(x)):
            ds = vmf_lookup(a, differential_system=True).get("differential_system")
            if not ds:
                continue
            elem = ds.get("df") if is_vf_case else ds.get("vf")
            if elem is None:
                continue
            basis_elems.append(elem)

        if not basis_elems:
            raise TypeError(
                "`annihilator` could not obtain a coordinate basis from VMF with differential objects. "
                "Suggestion: initialize coordinates with `createVariables(..., withVF=True)` (or `createVariables(..., complex=True)` as appropriate)."
            )

    if not coherent_coordinates_checked:
        objList = [allToReal(o) for o in objList]
        basis_elems = [allToReal(e) for e in basis_elems]

    n = len(basis_elems)
    label = create_key(prefix="var")
    vars_list = [symbol(f"{label}{i}") for i in range(n)]

    general = sum(u * e for u, e in zip(vars_list, basis_elems))

    if is_vf_case:
        eqns = [general(vf) for vf in objList]
    else:
        if expensive_path_needed:
            eqns = []
            combos = dict()
            for df in objList:
                deg = df.max_degree
                if deg == 0:
                    eqns.append(df)
                else:
                    if deg not in combos:
                        combos[deg] = chooseOp(
                            range(n),
                            deg - 1,
                            withOrder=True,
                            withoutReplacement=True,
                        )
                    for tail in combos[deg]:
                        trailing = [basis_elems[idx] for idx in tail]
                        eqns += [df(general, *trailing) for df in objList]
        else:
            eqns = [df(general) for df in objList]

    scalar_eqns = []
    for e in eqns:
        if get_dgcv_category(e) == "tensor_field":
            scalar_eqns.extend(e.coeff_dict.values())
        else:
            scalar_eqns.append(e)

    sols = solve_dgcv(scalar_eqns, vars_list, method="linsolve")
    if not sols:
        if _pass_error_report == retrieve_passkey():
            return "`annihilator` found no solutions."
        return []

    sol = sols[0]

    def _apply_sol(expr, sol_dict, scale=False):
        sol = subs(expr, sol_dict) if isinstance(sol_dict, dict) else expr
        if scale:
            f = getattr(sol, "scale_to_polynomial_attempt", None)
            sol = f() if callable(f) else sol
        return sol

    general_solution = _apply_sol(general, sol)

    free_vars = [u for u in vars_list if sol.get(u, u) == u]

    if not free_vars:
        if polynomial_bases:
            f = getattr(general_solution, "scale_to_polynomial_attempt", None)
            general_solution = f() if callable(f) else general_solution
        return [general_solution]

    out = []
    for v in free_vars:
        assign = {u: 0 for u in free_vars}
        assign[v] = 1
        out.append(_apply_sol(general_solution, assign, scale=polynomial_bases))
    out = [elem for elem in out if not _scalar_is_zero(elem)]
    return out



"""
package: dgcv - Differential Geometry with Complex Variables
module: CR_geometry

Description: This module provides tools specific to CR (Cauchy-Riemann) geometry within the dgcv package.
It includes functions for constructing CR hypersurfaces and  computing symmetries.

Key classes:
    - CR_structure
    - abstract_CR_structure

Key Functions:
    - tangencyObstruction(): Computes the tangency obstruction for a holomorphic vector field's
      real part to be tangent to a CR hypersurface.
    - weightedHomogeneousVF(): Creates a general weighted homogeneous vector field in a
      specified coordinate space.
    - findWeightedCRSymmetries(): Attempts to find infinitesimal symmetries of a weighted CR
      hypersurface.
    - model2Nondegenerate(): Builds the defining equation for a 2-nondegenerate model
      hypersurface.

Author (of this module): David Sykes (https://realandimaginary.com/dgcv/)

License:
    MIT License
"""

# -----------------------------------------------------------------------------
# imports and broadcasting
# -----------------------------------------------------------------------------
from __future__ import annotations

from typing import Any, Iterable, List, Literal, Sequence, Tuple

from ._config import dgcv_warning
from ._safeguards import (
    create_key,
    query_dgcv_categories,
    retrieve_passkey,
)
from .arrays import matrix_dgcv
from .backends import simplify_dgcv
from .backends._engine import engine_module
from .backends._symbolic_router import (
    _scalar_is_zero,
    as_numer_denom,
    cancel,
    conjugate,
    expand,
    get_free_symbols,
    subs,
)
from .backends._types_and_constants import (
    expr_numeric_types,
    imag_unit,
    rational,
    symbol,
)
from .base import dgcv_class
from .complex_structures import Del, DelBar
from .conversions import allToReal, allToSym, realToHol, realToSym, symToHol
from .dgcv_core import (
    complex_struct_op,
    holVF_coeffs,
    polynomial_dgcv,
    realPartOfVF,
    tensor_field_class,
    variableProcedure,
    wedge,
)
from .filtered_structures import distribution
from .polynomials import createMultigradedPolynomial
from .solvers import solve_dgcv
from .vector_fields_and_differential_forms import (
    LieDerivative,
    annihilator,
    exteriorDerivative,
    get_VF,
)
from .vmf import clearVar, listVar, order_coordinates, vmf_lookup

__all__ = [
    "CR_structure",
    "abstract_CR_structure",
    "tangencyObstruction",
    "weightedHomogeneousVF",
    "findWeightedCRSymmetries",
    "model2Nondegenerate",
]


# -----------------------------------------------------------------------------
# classes
# -----------------------------------------------------------------------------
class CR_structure(dgcv_class):
    """
    Represents local CR structure on a real submanifold in complex space given locally by defining equations.
    """

    def __init__(
        self,
        holomorphic_coordinates: list | tuple,
        defining_equations: list | tuple | dict,
        default_coordinate_format: Literal["real", "complex"] = None,
        parameters: set | list | tuple = None,
        enforce_finding_real_tangent_bundle_basis: bool = False,
        weights: list | tuple = None,
    ):
        self._weight_map = None
        if isinstance(weights, (list, tuple)):
            if len(weights) != len(holomorphic_coordinates):
                dgcv_warning(
                    "The provided `weights` list is not the same length as the provided holomorphic coordinates list, and was ignored."
                )
            else:
                self._weight_map = {
                    v: w for v, w in zip(holomorphic_coordinates, weights)
                }
        formatted_coordinates = []
        parameters = set() if parameters is None else set(parameters)
        seen_coordinates = set()
        omitted = set()
        for var in holomorphic_coordinates:
            holoVar = (
                vmf_lookup(var, relatives=True)
                .get("relatives", dict())
                .get("holo", None)
            )
            if holoVar is None:
                omitted.add(holoVar)
            elif holoVar not in seen_coordinates:
                formatted_coordinates.append(holoVar)
                seen_coordinates.add(holoVar)
        if len(omitted) > 0:
            dgcv_warning(
                "Some of the elements provided in `holomorphic_coordinates` are not part of a complex coordinate system registered in the dgcv VMF. They have been omitted from the CR structure's coordinate system, and regarded as structure parameters instead."
            )
            parameters |= omitted
        self.default_coordinate_format = {None: "complex", "real": "real"}.get(
            default_coordinate_format, "complex"
        )
        self.graph_format = False

        if isinstance(defining_equations, dict):
            self.graph_format = True
            eqns_data = []
            ri_types = {"real", "imag"}
            graph_coordinates = set()
            added_vars = set()
            for k, v in defining_equations.items():
                real_k = allToReal(k)
                key_info = vmf_lookup(real_k, relatives=True)
                if key_info.get("type", None) != "coordinate":
                    self.graph_format = False
                    dgcv_warning(
                        "Initializing a `CR_structure` with `defining_equations` in the dictionary format is intended for inputing equations in a graph format (e.g., graph coordinates equal functions of other coordinates). To process it as `graph form`, the dictionary's keys must belong to a coordinate system registered in the dgcv VMF. The given dictionary fails this condition, so preserving `graph form` has been abandoned. Tip: use `createVariables` to build such coordinate systems."
                    )
                    break
                if key_info.get("sub_type", None) not in ri_types:
                    self.graph_format = False
                    raise (
                        "Initializing a `CR_structure` with `defining_equations` in the dictionary format is intended for inputing equations in a graph format (e.g., graph coordinates equal functions of other coordinates). To process it as `graph form`, the dictionary's keys must be real or imaginary parts of holomorphic coordinates belonging to a coordinate system registered in the dgcv VMF. The given dictionary fails this condition, so preserving `graph form` has been abandoned."
                    )
                    break
                holoVar = key_info.get("relatives", dict()).get("holo", None)
                if holoVar not in seen_coordinates:
                    added_vars.add(holoVar)
                    seen_coordinates.add(holoVar)
                    formatted_coordinates.append(holoVar)

                real_v = allToReal(v)
                v_fv = get_free_symbols(v)
                if real_k in v_fv:
                    self.graph_format = False
                    dgcv_warning(
                        "Initializing a `CR_structure` with `defining_equations` in the dictionary format is intended for inputing equations in a graph format (e.g., graph coordinates equal functions of other coordinates). The given dictionary however has an equation of the form x=f(x,y,...), so preserving `graph form` has been abandoned."
                    )
                    break
                graph_coordinates.add(real_k)
                conv_v = (
                    real_v
                    if self.default_coordinate_format == "real"
                    else realToSym(real_v)
                )
                conv_k = (
                    real_k
                    if self.default_coordinate_format == "real"
                    else realToSym(real_k)
                )
                eqns_data.append(
                    {
                        "graph_coor": real_k,
                        "graph_function": conv_v,
                        "conv_graph_coor": conv_k,
                        "real_graph_fun": real_v,
                        "fun_symbols": v_fv,
                    }
                )
            if self.graph_format is True:
                if len(added_vars) > 0:
                    dgcv_warning(
                        f"Some of the given graph coordinates come from holomorphic coordinates not present in the given `holomorphic_coordinates` list, so new coordinates {added_vars} have been added to it."
                    )

                def sort_key(elem):
                    return len(
                        [var for var in graph_coordinates if var in elem["fun_symbols"]]
                    )

                eqns_data = sorted(eqns_data, key=sort_key)
                self.flattened_defining_equations = tuple(
                    [
                        eqn_data["graph_function"] - eqn_data["conv_graph_coor"]
                        for eqn_data in eqns_data
                    ]
                )
                self.graph_equations = {
                    eqn_data["graph_coor"]: eqn_data["graph_function"]
                    for eqn_data in eqns_data
                }
                param_extraction = set()
                for eqn in eqns_data:
                    for var in eqn["fun_symbols"]:
                        holoVar = (
                            vmf_lookup(var, relatives=True)
                            .get("relatives", dict())
                            .get("holo", None)
                        )
                        if holoVar not in seen_coordinates:
                            if var not in parameters:
                                param_extraction.add(var)
                                parameters.add(var)
                if len(param_extraction) > 0:
                    dgcv_warning(
                        f"The given equations include the variables {param_extraction}, which are not present in the given coordinate or parameter sets. They have been added to the structure parameters."
                    )
                self.parameters = parameters
                self.holomorphic_coordinates = tuple(
                    order_coordinates(formatted_coordinates)
                )
            else:
                try:
                    defining_equations = [v - k for k, v in defining_equations.items()]
                except Exception:
                    raise (
                        "Given `defining_equations` parameter is in an unsuported format."
                    )
        if self.graph_format is False:
            if isinstance(defining_equations, expr_numeric_types()):
                defining_equations = [defining_equations]
            if not isinstance(defining_equations, (list, tuple)):
                raise (
                    "Given `defining_equations` parameter is in an unsuported format."
                )
            self.flattened_defining_equations = (
                [allToReal(eqn) for eqn in defining_equations]
                if self.default_coordinate_format == "real"
                else [allToSym(eqn) for eqn in defining_equations]
            )
            param_extraction = set()
            for eqn in self.flattened_defining_equations:
                for var in get_free_symbols(eqn):
                    holoVar = (
                        vmf_lookup(var, relatives=True)
                        .get("relatives", dict())
                        .get("holo", None)
                    )
                    if holoVar not in seen_coordinates:
                        if var not in parameters:
                            param_extraction.add(var)
                            parameters.add(var)
            if len(param_extraction) > 0:
                dgcv_warning(
                    f"The given equations include the variables {param_extraction}, which are not present in the given coordinate or parameter sets. They have been added to the structure parameters."
                )
            self.parameters = parameters
            self.holomorphic_coordinates = tuple(
                order_coordinates(formatted_coordinates)
            )

            self.graph_equations = None
        self._efrtbb = enforce_finding_real_tangent_bundle_basis
        self.CR_codimension = len(self.flattened_defining_equations)
        self._CR_distribution = None
        self._holomorphic_CR_distribution = None
        self._tangent_bundle = None
        self._real_tangent_bundle = None
        self._antiholomorphic_CR_dist = None
        self._antiholomorphic_coordinates = None
        self._real_part_coordinates = None
        self._imag_part_coordinates = None
        self._real_coordinates = None
        self._complex_coordinates = None
        self._levi_2form = None
        self._graph_format_cache = None
        self._LFM = None
        self._Freeman_filtration = None
        self._real_def_eqns_cache = None
        self._symmetries = dict()

        super().__init__()

    @property
    def _graph_format_restriction(self):
        if self._graph_format_cache is None:
            if not self.graph_format:
                self._graph_format_cache = dict()
            else:
                if self.default_coordinate_format == "real":
                    self._graph_format_cache = self.graph_equations
                else:
                    self._graph_format_cache = {
                        k: allToReal(
                            v,
                            convert_everything=False,
                            variables_scope=list(self.graph_equations.keys()),
                        )
                        for k, v in self.graph_equations.items()
                    }
        return self._graph_format_cache

    @property
    def _real_defining_equations(self):
        if self._real_def_eqns_cache is None:
            self._real_def_eqns_cache = (
                self.flattened_defining_equations
                if self.default_coordinate_format == "real"
                else [allToReal(eqn) for eqn in self.flattened_defining_equations]
            )
        return self._real_def_eqns_cache

    @property
    def CR_dimension(self):
        return self.CR_distribution.rank // 2

    @property
    def CR_distribution(self):
        if self._CR_distribution is None:
            t_basis = self.tangent_bundle.vf_basis
            canonf = self.tangent_bundle.canonical_form
            Jt_basis = [complex_struct_op(vf) for vf in t_basis]
            vl = create_key("var")
            variables = [symbol(f"{vl}{idx}") for idx in range(len(Jt_basis))]
            gen_elem = sum(v * elem for v, elem in zip(variables, Jt_basis))
            obst = wedge(gen_elem, canonf)
            sol = solve_dgcv(obst, variables, method="linsolve")
            gen_sol = subs(gen_elem, sol[0])
            free_vars = set()
            for expr in sol[0].values():
                free_vars |= {var for var in get_free_symbols(expr) if var in variables}
            zeroing = {var: 0 for var in free_vars}
            self._CR_distribution = distribution(
                [subs(gen_sol, zeroing | {var: 1}) for var in free_vars],
                _assume_minimal_Data=True,
            )
        return self._CR_distribution

    @property
    def holomorphic_CR_distribution(self):
        if self._holomorphic_CR_distribution is None:
            if self._antiholomorphic_CR_dist is None:
                I = imag_unit()
                self._holomorphic_CR_distribution = distribution(
                    [
                        X - I * complex_struct_op(X)
                        for X in self.CR_distribution.vf_basis
                    ],
                    find_basis=True,
                    find_polynomial_spanners=True,
                    dimension_hint=self.CR_dimension,
                )
            else:
                self._holomorphic_CR_distribution = conjugate(
                    self._antiholomorphic_CR_dist, symbolic=True
                )
        return self._holomorphic_CR_distribution

    @property
    def antiholomorphic_CR_distribution(self):
        if self._antiholomorphic_CR_dist is None:
            if self._holomorphic_CR_distribution is None:
                I = imag_unit()
                self._antiholomorphic_CR_dist = distribution(
                    [
                        X + I * complex_struct_op(X)
                        for X in self.CR_distribution.vf_basis
                    ],
                    find_basis=True,
                    find_polynomial_spanners=True,
                    dimension_hint=self.CR_dimension,
                )
            else:
                self._antiholomorphic_CR_dist = conjugate(
                    self._holomorphic_CR_distribution, symbolic=True
                )
        return self._antiholomorphic_CR_dist

    @property
    def real_coordinates(self):
        if self._real_coordinates is None:
            if self._real_part_coordinates is None:
                self._real_part_coordinates = tuple(
                    [
                        vmf_lookup(var, relatives=True)["relatives"]["real"]
                        for var in self.holomorphic_coordinates
                    ]
                )
            if self._imag_part_coordinates is None:
                self._imag_part_coordinates = tuple(
                    [
                        vmf_lookup(var, relatives=True)["relatives"]["imag"]
                        for var in self.holomorphic_coordinates
                    ]
                )
            self._real_coordinates = (
                self._real_part_coordinates + self._imag_part_coordinates
            )
        return self._real_coordinates

    @property
    def tangent_bundle(self):
        if self._tangent_bundle is None:
            if self._efrtbb is True:
                defeqns = self._real_defining_equations
                coor = self.real_coordinates
            else:
                defeqns = self.flattened_defining_equations
                coor = self.default_coordinates
            d_rho_list = [exteriorDerivative(eqn) for eqn in defeqns]
            basis = annihilator(
                d_rho_list,
                coor,
                polynomial_bases=True,
                coherent_coordinates_checked=True,
            )
            if self._efrtbb is True and self.default_coordinate_format != "real":
                basis = [realToSym(arg) for arg in basis]
            self._tangent_bundle = distribution(
                basis,
                formatting=self.default_coordinate_format,
                _assume_minimal_Data=True,
            )
        return self._tangent_bundle

    @property
    def real_tangent_bundle(self):
        if self._real_tangent_bundle is None:
            I, tangent = imag_unit(), self.tangent_bundle.vf_basis
            self._real_tangent_bundle = (
                self.tangent_bundle
                if self.default_coordinate_format == "real"
                else distribution(
                    [X + conjugate(X, symbolic=True) for X in tangent]
                    + [I * (conjugate(X, symbolic=True) - X) for X in tangent],
                    find_basis=True,
                )
            )
        return self._real_tangent_bundle

    @property
    def complex_coordinates(self):
        if self._complex_coordinates is None:
            if self._antiholomorphic_coordinates is None:
                self._antiholomorphic_coordinates = tuple(
                    [
                        vmf_lookup(var, relatives=True)["relatives"]["anti"]
                        for var in self.holomorphic_coordinates
                    ]
                )
            self._complex_coordinates = (
                self.holomorphic_coordinates + self._antiholomorphic_coordinates
            )
        return self._complex_coordinates

    @property
    def default_coordinates(self):
        if self.default_coordinate_format == "real":
            return self.real_coordinates
        return self.complex_coordinates

    def Levi_form_skew_unrestricted(
        self, vf1, vf2, format_scalar_as_vector=False, skip_fast_simplify=False
    ):
        """
        Evaluates a representative of the Levi form on a pair of vector fields.
        Intended for applying to pairs of vector fields from the real CR
        distribution or its complexification (i.e., from the holomorphic +
        antiholomorphic CR distributions).

        By default the returned value is either a 1-by-n array when the CR
        codimension is n>2, or a scalar when the CR codimension is 1. To make
        the codimension 1 case also return an array for consistent formatting
        set the keyword `format_scalar_as_vector=True`.

        Caution/note: Intended only for applying to pairs of vector fields from
        the complexified CR distribution. It will not raise errors or warnings
        if applied to other vector fields however.
        """
        if self._levi_2form is None:
            self._levi_2form = matrix_dgcv(
                [Del(DelBar(rho)) for rho in self.flattened_defining_equations]
            )
        out = self._levi_2form.apply(lambda x: x.__call__(vf1, vf2))
        if skip_fast_simplify is False:
            out = expand(out)
        if self.CR_codimension == 1 and not format_scalar_as_vector:
            out = out[0]
        return out

    def Levi_form_skew(self, vf1, vf2, format_scalar_as_vector=False):
        """
        Evaluates a representative of the Levi form on a pair of vector fields.
        Intended for applying to pairs of vector fields from the real CR
        distribution or its complexification (i.e., from the holomorphic +
        antiholomorphic CR distributions). This Levi form is that of the
        distribution, in the usual sense, rather than the sesquilinear Levi form
        of CR geometry defined with conjugation hooked into an argument.

        By default the returned value is either a 1-by-n array when the CR
        codimension is n>2, or a scalar when the CR codimension is 1. To make
        the codimension 1 case also return an array for consistent formatting
        set the keyword `format_scalar_as_vector=True`.

        Caution/note: Intended only for applying to pairs of vector fields from
        the complexified CR distribution. It will not raise errors or warnings
        if applied to other vector fields however.
        """
        if self.graph_format is False:
            dgcv_warning(
                "This CR structure was not set up in `graph format`, so the computed Levi form value has not been restricted to the submanifold. Suggestion: provide structure data in `graph format` or use `Levi_form_skew_unrestricted` instead."
            )
        return subs(self.Levi_form_skew_unrestricted, self._graph_format_restriction)

    def Levi_form_unrestricted(self, vf1, vf2, format_scalar_as_vector=False):
        """
        Evaluates a representative of the Levi form on a pair of vector fields.
        Intended for applying to pairs of vector fields from the holomorphic CR
        distribution.

        By default the returned value is either a 1-by-n array when the CR
        codimension is n>2, or a scalar when the CR codimension is 1. To make
        the codimension 1 case also return an array for consistent formatting
        set the keyword `format_scalar_as_vector=True`.

        Caution/note: Intended only for applying to pairs of vector fields from
        the holomorphic CR distribution. It will not raise errors or warnings
        if applied to other vector fields however.
        """
        return (
            imag_unit()
            / 2
            * self.Levi_form_skew_unrestricted(
                vf1,
                conjugate(vf2, symbolic=True),
                format_scalar_as_vector=format_scalar_as_vector,
            )
        )

    def Levi_form(self, vf1, vf2, format_scalar_as_vector=False):
        """
        Evaluates a representative of the Levi form on a pair of vector fields.
        This Levi form is the usual CR geometry one, i.e.,
        sesquilinear with complex conjugation hooked into the ordinary skew
        symmetric Levi form of the complexified CR distribution.

        By default the returned value is either a 1-by-n array when the CR
        codimension is n>2, or a scalar when the CR codimension is 1. To make
        the codimension 1 case also return an array for consistent formatting
        set the keyword `format_scalar_as_vector=True`.

        Caution/note: Intended only for applying to pairs of vector fields from
        the holomorphic CR distribution. It will not raise errors or warnings
        if applied to other vector fields however.
        """
        if self.graph_format is False:
            dgcv_warning(
                "This CR structure was not set up in `graph format`, so the computed Levi form value has not been restricted to the submanifold. Suggestion: provide structure data in `graph format` or use `Levi_form_unrestricted` instead."
            )
        return subs(
            self.Levi_form_unrestricted(vf1, vf2), self._graph_format_restriction
        )

    @property
    def Levi_form_matrix(self):
        if self._LFM is None:
            vf_basis = self.holomorphic_CR_distribution.vf_basis

            def entry(i, j):
                return self.Levi_form(vf_basis[i], vf_basis[j])

            self._LFM = matrix_dgcv(
                shape=(self.CR_dimension, self.CR_dimension), entry_rule=entry
            )
        return self._LFM

    @property
    def Freeman_filtration(self):
        if self._Freeman_filtration is None:
            antihol = self.antiholomorphic_CR_distribution
            levels = [self.holomorphic_CR_distribution.vf_basis]
            vl = create_key("var")
            local_vars = [symbol(f"{vl}{idx}") for idx in range(len(levels[0]))]

            def descent(level_list, pair, variables):
                if len(level_list[0]) == 0:
                    return level_list
                level = level_list[0]
                v_trunc = variables[: len(level)]
                general_elem = sum(v * elem for v, elem in zip(v_trunc, level))
                new_c_form = wedge(pair.canonical_form, *level)
                eqns = [
                    wedge(LieDerivative(general_elem, vf), new_c_form)
                    for vf in pair.vf_basis
                ]
                solution = solve_dgcv(eqns, v_trunc, method="linsolve")
                gen_solution = subs(general_elem, solution[0])
                free_vars, var_set = set(), set(v_trunc)
                for expr in solution[0].values():
                    free_vars |= {
                        var for var in get_free_symbols(expr) if var in var_set
                    }
                zeroing = {var: 0 for var in free_vars}
                new_level = [
                    [
                        subs(
                            gen_solution, zeroing | {var: 1}
                        ).scale_to_polynomial_attempt()
                        for var in free_vars
                    ]
                ]
                lnl = len(new_level[0])
                if len(level_list[0]) == lnl:
                    return level_list
                if lnl == 0:
                    return [[]] + level_list
                level_list = new_level + level_list
                return descent(level_list, pair, variables)

            filtration_bases = descent(levels, antihol, local_vars)
            alligned_bases = filtration_bases[:1]
            if len(filtration_bases) > 1:
                obstruction = wedge(*filtration_bases[0])
                if obstruction is None:
                    obstruction = tensor_field_class(coeff_dict={tuple(): 1})
                dim = len(filtration_bases[0])
                for basis in filtration_bases[1:]:
                    current_basis = list(alligned_bases[0])
                    new_dim = len(basis)
                    discrep = new_dim - dim
                    dim = new_dim
                    for elem in basis:
                        if discrep == 0:
                            break
                        test = simplify_dgcv(wedge(elem, obstruction))
                        if _scalar_is_zero(test):
                            continue
                        discrep += -1
                        current_basis.append(elem)
                        obstruction = test
                    alligned_bases = [current_basis] + alligned_bases
            self._Freeman_filtration = [
                distribution(
                    base,
                    assume_compatibility=True,
                    formatting=self.default_coordinate_format,
                    _assume_minimal_Data=True,
                )
                for base in alligned_bases
            ]

        return self._Freeman_filtration

    @property
    def nondegeneracy_order(self):
        if len(self.Freeman_filtration[-1].vf_basis) > 0:
            return "infinity"
        return len(self.Freeman_filtration) - 1

    def compute_weighted_symmetries(
        self,
        target_weights,
        coordinate_weights=None,
        report_progress=False,
        verbose=False,
        degreeCap: int = 0,
        assume_polynomial: bool = False,
        simplify: bool = False,
    ):
        if not self.graph_format:
            raise RuntimeError(
                "`CR_structure.compute_weighted_symmetries` is only supported for `CR_structure` instances initialized in a 'graph format'. The `defining_equations` parameter given for instance initialization needs to have been in graph format, meaning as a dictionary whose keys are graphing variables and whose values are functions of other variables."
            )
        if isinstance(target_weights, expr_numeric_types()):
            target_weights = [target_weights]
        else:
            target_weights = list(target_weights)
        if not all(isinstance(elem, expr_numeric_types()) for elem in target_weights):
            dgcv_warning(
                "provided `target_weights` to compute symmetries for is not in a supported format. Should be either a simple scalar like an integer or rational or a list of them."
            )
            return {}
        caching = False
        if coordinate_weights is None:
            caching = True
            if self._weight_map is None:
                raise RuntimeError(
                    "This `CR_structure` instance was not initialized with valid coordinate weights so the optional `coordinate_weights` keyword needs to be provided instead."
                )
            else:
                coordinate_weights = [
                    self._weight_map[var]
                    for var in self.holomorphic_coordinates
                    if var in self.holomorphic_coordinates
                ]
            if len(coordinate_weights) != len(self.holomorphic_coordinates):
                raise RuntimeError(
                    "This `CR_structure` instance was not initialized with valid coordinate weights so the optional `coordinate_weights` keyword needs to be provided instead. A possible reason could be that the given `holomorphic_coordinates` list in the `CR_structure` initialization was incomplete, with additional variables appearing in the initialization defining equation system."
                )
        elif len(coordinate_weights) != len(self.holomorphic_coordinates):
            raise RuntimeError(
                "`coordinate_weights` must be an integer list of the same length as the CR_structure.holomorphic_coordinates list."
            )
        symmetries = dict()
        if verbose:
            print("Using variable weights:")
            print(
                f"{', '.join(['[' + str(var) + '] = ' + str(w) for var, w in zip(self.holomorphic_coordinates, coordinate_weights)])}"
            )
            print(" ")
        for weight in target_weights:
            if report_progress:
                print(f"Computing symmetries of weight {weight}:")
            if weight in self._symmetries and caching is True:
                symmetries[weight] = self._symmetries[weight]
            else:
                symmetries[weight] = findWeightedCRSymmetries(
                    self.flattened_defining_equations,
                    holomorphic_coordinates=self.holomorphic_coordinates,
                    coordinate_weights=coordinate_weights,
                    symmetry_weight=weight,
                    graph_variable=list(self.graph_equations.keys()),
                    returnVectorFieldBasis=True,
                    degreeCap=degreeCap,
                    assume_polynomial=assume_polynomial,
                    simplify=simplify,
                    parameters=self.parameters if self.parameters else None,
                )
                if caching is True:
                    self._symmetries[weight] = symmetries[weight]
            if report_progress:
                number_found = len(symmetries[weight])
                grammar_detail = (
                    ["symmetry", "was"] if number_found == 1 else ["symmetries", "were"]
                )
                print(
                    f"{number_found} {grammar_detail[0]} of weight {weight} {grammar_detail[1]} found."
                )
                print(" ")
        return symmetries


class abstract_CR_structure(dgcv_class):
    """
    Represents local CR structure in terms of real coordinates. Initialization data
    should be the holomorphic component of the complexified CR distribution,
    i.e., the i-eigenspace bundle of the CR complex structure operator.
    """

    def __init__(self, holomorphic_distribution, validate_structure_data=False):
        super().__init__()


# -----------------------------------------------------------------------------
# CR tools
# -----------------------------------------------------------------------------
def tangencyObstruction(
    vf,
    CR_defining_expr,
    graph_variable,
    simplify: bool = False,
    data_already_in_real_coor: bool = False,
    *args,
):
    CR_defs = _as_list(CR_defining_expr)
    graph_vars = _as_list(graph_variable)

    if len(CR_defs) != len(graph_vars):
        raise ValueError(
            "`graph_variable` and `CR_defining_expr` must have the same length."
        )

    if not data_already_in_real_coor:
        graph_vars = [allToReal(v) for v in graph_vars]
        CR_defs = [allToReal(f) for f in CR_defs]
        vf = allToReal(vf)

    rVF = realPartOfVF(vf)
    real_eval = [rVF(gv - f) for gv, f in zip(graph_vars, CR_defs)]

    mapping = dict(zip(graph_vars, CR_defs))
    out = [subs(expr, mapping) for expr in real_eval]

    if simplify:
        out = [simplify_dgcv(x) for x in out]

    return out


def weightedHomogeneousVF(
    varSpace,
    weight,
    weights,
    varLabel,
    degreeCap: int = 0,
    _tempVar=None,
    assumeReal=None,
):
    if not _is_sequence(weight):
        weight = [weight]
    if not weights or not _is_sequence(weights[0]):
        weights = [weights]

    weight_list = list(weight)
    weight_systems = list(weights)

    if not all(len(ws) == len(varSpace) for ws in weight_systems):
        raise KeyError(
            "`weightedHomogeneousVF` expects each weight system to match len(varSpace)."
        )
    if len(weight_list) != len(weight_systems):
        raise KeyError(
            "`weightedHomogeneousVF` expects len(weight) == number of weight systems."
        )

    polys = []
    for j in range(len(varSpace)):
        degs = [d + ws[j] for d, ws in zip(weight_list, weight_systems)]
        polys.append(
            createMultigradedPolynomial(
                f"{varLabel}_{j}_",
                degs,
                varSpace,
                weight_systems,
                degreeCap=degreeCap,
                _tempVar=_tempVar,
                assumeReal=assumeReal,
                report_vmf_updates=False,
            )
        )

    Dz_list = get_VF(*varSpace)
    out = None
    for p, Dz in zip(polys, Dz_list):
        if not p:
            continue
        term = p * Dz
        out = term if out is None else (out + term)

    if out is None:
        out = type(Dz_list[0])(coeff_dict={tuple(): 0}, data_shape="all")

    return out


def findWeightedCRSymmetries(
    graph_function,
    holomorphic_coordinates,
    coordinate_weights,
    symmetry_weight,
    graph_variable,
    coeff_label=None,
    degreeCap: int = 0,
    returnVectorFieldBasis: bool = False,
    returnAllformats: bool = False,
    simplifyingFactor=None,
    assume_polynomial: bool = False,
    simplify: bool = False,
    parameters=None,
):
    if not coordinate_weights or not _is_sequence(coordinate_weights[0]):
        coordinate_weights = [coordinate_weights]
    if not _is_sequence(symmetry_weight):
        symmetry_weight = [symmetry_weight]

    weight_systems = list(coordinate_weights)
    sym_weight_list = list(symmetry_weight)

    if len(sym_weight_list) != len(weight_systems):
        raise KeyError(
            "`findWeightedCRSymmetries` requires one symmetry weight per weight system."
        )

    if returnAllformats:
        returnVectorFieldBasis = True

    def _coord_ri_from_vmf(coords):
        out = set()
        for v in coords:
            info = vmf_lookup(v, relatives=True)
            if info.get("type") != "coordinate":
                continue
            rel = info.get("relatives") or {}
            r = rel.get("real", None)
            i = rel.get("imag", None)
            if r is not None:
                out.add(r)
            if i is not None:
                out.add(i)
            if r is None and i is None:
                out.add(v)
        return out

    vfA = weightedHomogeneousVF(
        holomorphic_coordinates,
        sym_weight_list,
        weight_systems,
        "ALoc",
        _tempVar=retrieve_passkey(),
        degreeCap=degreeCap,
        assumeReal=True,
    )
    vfB = weightedHomogeneousVF(
        holomorphic_coordinates,
        sym_weight_list,
        weight_systems,
        "BLoc",
        _tempVar=retrieve_passkey(),
        degreeCap=degreeCap,
        assumeReal=True,
    )

    vf_candidate_holo = vfA + imag_unit() * vfB

    tan_obst_list = tangencyObstruction(
        vf_candidate_holo,
        graph_function,
        graph_variable,
        simplify=simplify,
    )

    free = get_free_symbols
    unwrap = _unwrap_symbolic

    var_all = set()
    for TO in tan_obst_list:
        var_all.update(free(unwrap(TO)))

    coord_ri = _coord_ri_from_vmf(holomorphic_coordinates)

    unknowns = set(var_all) - set(coord_ri)
    if parameters is not None:
        unknowns -= set(parameters)

    if coeff_label is None:
        coeff_label = create_key(prefix="X", key_length=5)

    unknowns_t = _stable_tuple(unknowns)

    graph_vars = _as_list(graph_variable)
    poly_gens_vars = _stable_tuple(list(coord_ri) + list(graph_vars))

    coef_eqns = []
    for TO in tan_obst_list:
        expr = unwrap(TO)
        expr = allToReal(expr)
        if not assume_polynomial:
            expr = as_numer_denom(cancel(expr))[0]
        expr = expand(expr)
        P = (
            expr
            if query_dgcv_categories(expr, "polynomial")
            else polynomial_dgcv(expr, poly_gens_vars)
        )
        coef_eqns.extend([allToReal(c) for c in P.get_coeffs(formatting="unformatted")])

    solutions = solve_dgcv(coef_eqns, unknowns_t, method="linsolve")

    if not solutions:
        if len(tan_obst_list) == 0 or all(
            _scalar_is_zero(unwrap(x)) for x in tan_obst_list
        ):
            clearVar(*listVar(temporary_only=True), report=False)
            if not returnVectorFieldBasis:
                return [0 for _ in holomorphic_coordinates]
            if returnAllformats:
                return [0 for _ in holomorphic_coordinates], []
            return []
        clearVar(*listVar(temporary_only=True), report=False)
        raise ValueError(
            "findWeightedCRSymmetries: no solution to the coefficient system."
        )

    sol = solutions[0] if isinstance(solutions, (list, tuple)) else solutions

    vf_solved = subs(vf_candidate_holo, sol)

    VFCLoc = holVF_coeffs(vf_solved, holomorphic_coordinates)

    sub_syms = set()
    for term in VFCLoc:
        sub_syms.update(free(unwrap(term)))

    sub_syms -= set(holomorphic_coordinates)
    sub_syms -= set(coord_ri)
    if parameters is not None:
        sub_syms.difference_update(parameters)

    if not sub_syms:
        clearVar(*listVar(temporary_only=True), report=False)
        if not returnVectorFieldBasis:
            return VFCLoc
        if returnAllformats:
            return VFCLoc, []
        return []

    sub_syms_t = _stable_tuple(sub_syms)

    created = variableProcedure(
        coeff_label,
        len(sub_syms_t),
        assumeReal=True,
        return_created_object=True,
    )
    coeff_vars = tuple(created[0]) if created else ()

    repl = dict(zip(sub_syms_t, coeff_vars))
    vf_param = subs(vf_solved, repl)

    clearVar(*listVar(temporary_only=True), report=False)

    if not returnVectorFieldBasis:
        return holVF_coeffs(vf_param, holomorphic_coordinates)

    zeroing = {w: 0 for w in coeff_vars}
    VF_basis = []
    for v in coeff_vars:
        d = dict(zeroing)
        d[v] = 1
        VF_basis.append(subs(vf_param, d))

    clearVar(coeff_label, report=False)

    if returnAllformats:
        return holVF_coeffs(vf_param, holomorphic_coordinates), VF_basis
    return VF_basis


def model2Nondegenerate(
    hermitian_matrix: Any,
    symmetric_matrix_function: Any,
    nondegenerate_coordinates: Sequence[Any],
    transverse_coordinate: Any,
    use_symbolic_conjugates: bool = False,
    return_matrices: bool = False,
    simplify: bool = True,
    return_CR_structure: bool = False,
    parameters: set = None,
    coordinates_to_weights_dict=None,
):
    """
    The parameter `coordinates_to_weights_dict` is relevant only if setting `return_CR_structure=True`. In this case it may be set to a dictionary with one key/value pair. The key should be a tuple of holomorphic coordinates, and the value should be a list/tuple of corresponding weights.
    """
    if return_CR_structure is True:
        use_symbolic_conjugates = True

    def _simp(x: Any) -> Any:
        return simplify_dgcv(x) if simplify else x

    def _as_md(mat: Any):
        return matrix_dgcv(mat)

    def _T(M: Any):
        try:
            return M.transpose()
        except Exception:
            pass
        fallback = M.to_symbolic_engine_class()
        try:
            transp = fallback.transpose
            if callable(transp):
                return transp()
            else:
                return transp
        except Exception:
            return fallback.T

    def _inv(M: Any):
        try:
            return M.inverse()
        except Exception:
            return M.symbolic_engine_method("__pow__", method_arguments=[-1])

    A = _as_md(hermitian_matrix)
    S = _as_md(symmetric_matrix_function)

    shp = getattr(A, "shape", None)
    if not (isinstance(shp, tuple) and len(shp) == 2 and shp[0] == shp[1]):
        raise TypeError("`hermitian_matrix` must be square.")
    n = int(shp[0])

    Iden = matrix_dgcv.eye(n)
    BARS = S.conjugate(symbolic=True)

    tc_info = vmf_lookup(transverse_coordinate, relatives=True)
    tc_holo = tc_info.get("relatives", dict()).get("holo")
    if tc_info.get("sub_type") == "holo" or tc_info.get("sub_type") == "anti":
        if use_symbolic_conjugates is False:
            transverse_coordinate = realToHol(
                tc_info.get("relatives", dict()).get("imag")
            )
        else:
            transverse_coordinate = realToSym(
                tc_info.get("relatives", dict()).get("imag")
            )
    elif tc_info.get("sub_type") not in {"real", "imag"}:
        dgcv_warning(
            "The given transverse coordinate is not associated with any complex coordinate system currently in the dgcv VMF."
        )

    base_coordinates = []
    seen = set()
    nonregistered_given = False
    transv_given = False
    minimal_not_given = False
    for var in nondegenerate_coordinates:
        holo = vmf_lookup(var, relatives=True).get("relatives", dict()).get("holo")
        if holo is not None and holo not in seen and tc_holo != holo:
            seen.add(holo)
            base_coordinates.append(holo)
        elif holo is None:
            nonregistered_given = True
        elif holo == tc_holo:
            transv_given = True
        else:
            minimal_not_given = True
    if nonregistered_given is True:
        dgcv_warning(
            "Some of the elements given in `nondegenerate_coordinates` are associated with a complex coordinate system in the dgcv VMF and were ignored."
        )
    if transv_given is True:
        dgcv_warning(
            "At least element in the given  `nondegenerate_coordinates` has the same holomorphic coordinate as the given transverse coordinate and was ignored."
        )
    if minimal_not_given is True:
        dgcv_warning(
            "Multiple elements given in `nondegenerate_coordinates` are associated with the same holomorphic coordinate and the repeats were ignored."
        )

    zVec = matrix_dgcv([[v] for v in base_coordinates])
    bzVec = zVec.conjugate(symbolic=True)

    half = rational(1, 2)

    At = _T(A)
    zt = _T(zVec)
    bzt = _T(bzVec)

    try:
        M1 = Iden - (BARS * At * S * A)
        M1_inv = _inv(M1)
        M1_inv_conj = M1_inv.conjugate(symbolic=True)
        M1_inv_conj_tran = M1_inv_conj.transpose()

        hFun = half * (A * M1_inv + M1_inv_conj_tran * A)
        sFun = A * M1_inv * BARS * At
        bsFun = At * M1_inv_conj * S * A
        expr = (zt * hFun * bzVec)[0, 0] + half * (
            (zt * sFun * zVec)[0, 0] + (bzt * bsFun * bzVec)[0, 0]
        )
    except Exception:
        raise RuntimeError(
            "`model2Nondegenerate` is unable to evaluate the model formula for the given parameters. Check that given matrices have compatible shapes."
        )
    graph = _simp(expr)
    if use_symbolic_conjugates is False:
        graph = symToHol(graph)
    result = graph - transverse_coordinate
    if return_CR_structure:
        kernel_total = get_free_symbols(S)
        seen = set()
        kernel_coordinates = []
        computed_parameters = set() if parameters is None else set(parameters)
        par_warning = False
        for var in kernel_total:
            if var not in computed_parameters:
                var_info = vmf_lookup(var, relatives=True).get("relatives", dict())
                holo = var_info.get("holo")
                if holo is None:
                    computed_parameters.add(var)
                    continue
                if (
                    var_info.get("real") in computed_parameters
                    or var_info.get("imag") in computed_parameters
                    or var_info.get("anti") in computed_parameters
                    or holo in computed_parameters
                ):
                    computed_parameters.add(var)
                    continue
                if var != holo:
                    par_warning is True
                if holo not in seen:
                    seen.add(holo)
                    kernel_coordinates.append(holo)
        if par_warning is True:
            dgcv_warning(
                "The provided symmetrix matrix has some complex coordinate elements other than holomorphic variables. They were not included among given parameters however. Since this matrix must be holomorphic in the underlying coordinates, you may have intended such elements to be included among the parameters. Use the optional `parameters` keyword to include them."
            )
        kernel_coordinates = order_coordinates(kernel_coordinates)
        holomorphic_coor = [tc_holo] + base_coordinates + kernel_coordinates
        if coordinates_to_weights_dict is not None:
            k, v = next(iter(coordinates_to_weights_dict.items()))
            cw_map = dict(zip(k, v))
            weights = [
                cw_map.get(var) for var in holomorphic_coor if var in holomorphic_coor
            ]
        else:
            weights = None
        return CR_structure(
            holomorphic_coor,
            {transverse_coordinate: graph},
            parameters=parameters,
            weights=weights,
        )
    if return_matrices:
        return result, hFun, sFun
    return result


# -----------------------------------------------------------------------------
# utilities
# -----------------------------------------------------------------------------
def _sym_engine_module():
    return engine_module()


def _is_sequence(x: Any) -> bool:
    return isinstance(x, (list, tuple))


def _as_list(x: Any) -> List[Any]:
    return list(x) if _is_sequence(x) else [x]


def _stable_tuple(seq: Iterable[Any]) -> Tuple[Any, ...]:
    out: List[Any] = []
    seen = set()
    for x in seq:
        if x in seen:
            continue
        seen.add(x)
        out.append(x)
    return tuple(out)


def _unwrap_symbolic(obj: Any) -> Any:
    if obj is None:
        return obj

    f = getattr(obj, "as_expr", None)
    if callable(f):
        try:
            return f()
        except TypeError:
            return f

    for attr in ("polyExpr", "_polyExpr", "expr", "_expr"):
        if hasattr(obj, attr):
            return getattr(obj, attr)

    P = getattr(obj, "poly_obj_unformatted", None)
    if P is not None and hasattr(P, "as_expr"):
        try:
            return P.as_expr()
        except Exception:
            pass

    return obj



"""
package: dgcv - Differential Geometry with Complex Variables
module: conversions

Author (of this module): David Sykes (https://realandimaginary.com/dgcv/)

License:
    MIT License
"""

# -----------------------------------------------------------------------------
# imports and broadcasting
# -----------------------------------------------------------------------------
from __future__ import annotations

import numbers
from typing import Any, Dict, Optional, Set, Tuple

from ._config import get_variable_registry
from ._safeguards import check_dgcv_category
from .backends._exact_arith import exact_reciprocal
from .backends._symbolic_router import get_free_symbols, simplify, subs
from .backends._types_and_constants import expr_types, imag_unit, is_atomic
from .combinatorics import carProd_with_weights_without_R, permSign
from .vmf import vmf_lookup

__all__ = [
    "allToHol",
    "allToReal",
    "allToSym",
    "holToReal",
    "holToSym",
    "symToHol",
    "symToReal",
    "realToHol",
    "realToSym",
]


# -----------------------------------------------------------------------------
# converters
# -----------------------------------------------------------------------------
def holToReal(
    expr,
    skipVar=None,
    convert_everything: bool = True,
    *,
    _conversion_dict=None,
    variables_scope=None,
):
    vr = get_variable_registry()
    conv = _get_conv_map(vr, "holToReal", skipVar, _conversion_dict, variables_scope)

    hooked = _dgcv_convert_hook(
        expr,
        "holToReal",
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
        conversion_dict=conv,
    )
    if hooked is not None:
        return hooked

    if check_dgcv_category(expr):
        new = _converter_hook(
            expr,
            to_kind="real",
            converter=holToReal,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
        if new is not None:
            return new

    return _convert_expr(expr, conv, convert_everything=convert_everything)


def realToSym(
    expr,
    skipVar=None,
    convert_everything: bool = True,
    *,
    _conversion_dict=None,
    variables_scope=None,
):
    vr = get_variable_registry()
    conv = _get_conv_map(vr, "realToSym", skipVar, _conversion_dict, variables_scope)

    hooked = _dgcv_convert_hook(
        expr,
        "realToSym",
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
        conversion_dict=conv,
    )
    if hooked is not None:
        return hooked

    if check_dgcv_category(expr):
        new = _converter_hook(
            expr,
            to_kind="complex",
            converter=realToSym,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
        if new is not None:
            return new

    return _convert_expr(expr, conv, convert_everything=convert_everything)


def symToHol(
    expr,
    skipVar=None,
    convert_everything: bool = True,
    *,
    _conversion_dict=None,
    variables_scope=None,
):
    vr = get_variable_registry()
    conv = _get_conv_map(vr, "symToHol", skipVar, _conversion_dict, variables_scope)

    hooked = _dgcv_convert_hook(
        expr,
        "symToHol",
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
        conversion_dict=conv,
    )
    if hooked is not None:
        return hooked

    if check_dgcv_category(expr):
        new = _converter_hook(
            expr,
            to_kind="complex",
            converter=symToHol,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
        if new is not None:
            return new

    return _convert_expr(expr, conv, convert_everything=convert_everything)


def realToHol(
    expr,
    skipVar=None,
    convert_everything: bool = True,
    *,
    _conversion_dict=None,
    variables_scope=None,
):
    vr = get_variable_registry()
    conv = _get_conv_map(vr, "realToHol", skipVar, _conversion_dict, variables_scope)

    hooked = _dgcv_convert_hook(
        expr,
        "realToHol",
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
        conversion_dict=conv,
    )
    if hooked is not None:
        return hooked

    if check_dgcv_category(expr):
        new = _converter_hook(
            expr,
            to_kind="complex",
            converter=realToHol,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
        if new is not None:
            return new

    return _convert_expr(expr, conv, convert_everything=convert_everything)


def symToReal(
    expr,
    skipVar=None,
    convert_everything: bool = True,
    *,
    _conversion_dict=None,
    variables_scope=None,
):
    vr = get_variable_registry()
    conv = _get_conv_map(vr, "symToReal", skipVar, _conversion_dict, variables_scope)

    hooked = _dgcv_convert_hook(
        expr,
        "symToReal",
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
        conversion_dict=conv,
    )
    if hooked is not None:
        return hooked

    if check_dgcv_category(expr):
        new = _converter_hook(
            expr,
            to_kind="real",
            converter=symToReal,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
        if new is not None:
            return new

    return _convert_expr(expr, conv, convert_everything=convert_everything)


def holToSym(expr, skipVar=None, convert_everything: bool = True, variables_scope=None):
    expr = holToReal(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )
    expr = realToSym(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )
    return expr


def allToReal(
    expr, skipVar=None, convert_everything: bool = True, variables_scope=None
):
    return symToReal(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )


def allToHol(expr, skipVar=None, convert_everything: bool = True, variables_scope=None):
    ###!!! refine later
    expr = symToHol(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )
    return realToHol(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )


def allToSym(expr, skipVar=None, convert_everything: bool = True, variables_scope=None):
    return holToSym(
        expr,
        skipVar=skipVar,
        convert_everything=convert_everything,
        variables_scope=variables_scope,
    )


def cleanUpConjugation(arg1, skipVar=None):
    return realToSym(
        symToReal(arg1, skipVar=skipVar, convert_everything=False),
        skipVar=skipVar,
        convert_everything=False,
    )


# -----------------------------------------------------------------------------
# utilities
# -----------------------------------------------------------------------------
categories = {"vector_field", "differential_form"}


def _maybe_to_real(expr, *, skipVar, variables_scope, convert_everything: bool):
    if not convert_everything:
        return None
    f = getattr(expr, "_to_real", None)
    if not callable(f):
        return None
    plan = {"skipVar": skipVar, "variables_scope": variables_scope}
    try:
        return f(plan=plan)
    except TypeError:
        return f(plan)


def _maybe_to_complex(expr, *, skipVar, variables_scope, convert_everything: bool):
    if not convert_everything:
        return None
    f = getattr(expr, "_to_complex", None)
    if not callable(f):
        return None
    plan = {"skipVar": skipVar, "variables_scope": variables_scope}
    try:
        return f(plan=plan)
    except TypeError:
        return f(plan)


def _maybe_apply(
    expr, converter, *, skipVar, variables_scope, convert_everything: bool
):
    ap = getattr(expr, "apply", None)
    if not callable(ap):
        return None

    def conv(e):
        return converter(
            e,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )

    return ap(conv)


def _converter_hook(
    expr,
    *,
    to_kind: str,  # "real" or "complex"
    converter,
    skipVar=None,
    variables_scope=None,
    convert_everything: bool = True,
):
    out = None

    if to_kind == "real":
        out = _maybe_to_real(
            expr,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )
    else:
        out = _maybe_to_complex(
            expr,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
        )

    base = out if out is not None else expr

    applied = _maybe_apply(
        base,
        converter,
        skipVar=skipVar,
        variables_scope=variables_scope,
        convert_everything=convert_everything,
    )
    if applied is not None:
        return applied

    if out is not None:
        return out

    return None


def _iter_skip_items(x) -> Tuple[Any, ...]:
    if x is None:
        return ()
    if isinstance(x, (str, bytes)):
        return (x,)
    try:
        return tuple(x)
    except TypeError:
        return (x,)


def _vmf_relatives(obj) -> Optional[dict]:
    try:
        info = vmf_lookup(obj, relatives=True)
    except Exception:
        return None
    if not isinstance(info, dict):
        return None
    rel = info.get("relatives", None)
    return rel if isinstance(rel, dict) else None


def _vmf_system_label(obj) -> Optional[str]:
    rel = _vmf_relatives(obj)
    if not rel:
        return None
    lab = rel.get("system_label", None)
    return lab if isinstance(lab, str) and lab else None


def _vmf_relative_atoms(obj) -> Optional[Tuple[Any, ...]]:
    rel = _vmf_relatives(obj)
    if not rel:
        return None
    atoms = []
    for k in ("standard", "holo", "anti", "real", "imag"):
        v = rel.get(k, None)
        if v is not None:
            atoms.append(v)
    return tuple(atoms) or None


def _prune_conversion_dict_for_skipVar(conv, vr, skipVar):
    items = _iter_skip_items(skipVar)
    if not items:
        return conv

    system_labels: Set[str] = set()
    atomic_items = []

    for it in items:
        if isinstance(it, str):
            system_labels.add(it)
        else:
            atomic_items.append(it)

    if system_labels:
        cvs = vr.get("complex_variable_systems", {})
        for lab in system_labels:
            cs = cvs.get(lab, {})
            family_values = cs.get("family_values", ((), (), (), ()))

            if cs.get("family_type") == "single":
                family_values = tuple((j,) for j in family_values)

            for block in family_values:
                for sym in block or ():
                    conv.pop(sym, None)

    for it in atomic_items:
        lab = _vmf_system_label(it)
        if lab and lab in system_labels:
            continue

        atoms = _vmf_relative_atoms(it)
        if not atoms:
            continue

        for sym in atoms:
            conv.pop(sym, None)

    return conv


def _scope_atoms(vr, variables_scope) -> Optional[Set[Any]]:
    items = _iter_skip_items(variables_scope)
    if not items:
        return None

    system_labels: Set[str] = set()
    atomic_items = []

    for it in items:
        if isinstance(it, str):
            system_labels.add(it)
        else:
            atomic_items.append(it)

    out: Set[Any] = set()

    if system_labels:
        cvs = vr.get("complex_variable_systems", {})
        for lab in system_labels:
            cs = cvs.get(lab, {})
            family_values = cs.get("family_values", ((), (), (), ()))

            if cs.get("family_type") == "single":
                family_values = tuple((j,) for j in family_values)

            for block in family_values:
                for sym in block or ():
                    out.add(sym)

    for it in atomic_items:
        lab = _vmf_system_label(it)
        if lab and lab in system_labels:
            continue

        atoms = _vmf_relative_atoms(it)
        if not atoms:
            continue

        out.update(atoms)

    return out


def _filter_conv_to_scope(base, allowed: Optional[Set[Any]]):
    if allowed is None:
        return base

    out = {}
    for k in allowed:
        try:
            v = base.get(k, None)
        except Exception:
            try:
                v = base[k]
            except Exception:
                v = None
        if v is not None:
            out[k] = v
    return out


def _dgcv_convert_hook(
    expr,
    conv,
    *,
    skipVar=None,
    variables_scope=None,
    convert_everything=True,
    conversion_dict=None,
):
    f = getattr(expr, "__dgcv_converter__", None)
    if not callable(f):
        return None

    try:
        return f(
            conv,
            skipVar=skipVar,
            variables_scope=variables_scope,
            convert_everything=convert_everything,
            conversion_dict=conversion_dict,
        )
    except TypeError:
        return f(
            conv,
            skipVar=skipVar,
            convert_everything=convert_everything,
            conversion_dict=conversion_dict,
        )


def _get_conv_map(vr, kind: str, skipVar, _conversion_dict, variables_scope):
    base = _conversion_dict
    if base is None:
        base = vr.get("conversion_dictionaries", {}).get(kind, {})

    allowed = _scope_atoms(vr, variables_scope)
    if allowed is not None:
        base = _filter_conv_to_scope(base, allowed)

    if not skipVar:
        return base

    try:
        conv = base.copy()
    except Exception:
        conv = dict(base)

    _prune_conversion_dict_for_skipVar(conv, vr, skipVar)
    return conv


def _filtered_subs(expr, conv) -> Optional[Dict[Any, Any]]:
    syms = get_free_symbols(expr)
    if not syms:
        return None

    out = {}
    for s in syms:
        try:
            v = conv[s]
        except Exception:
            try:
                v = conv.get(s, None)
            except Exception:
                v = None
        if v is not None:
            out[s] = v

    return out or None


def _convert_expr(expr, conv, *, convert_everything: bool):
    if isinstance(expr, numbers.Number):
        return expr

    if is_atomic(expr):
        try:
            return conv.get(expr, expr)
        except Exception:
            return expr

    if isinstance(expr, expr_types()):
        fs = _filtered_subs(expr, conv)
        if not fs:
            return expr
        res = subs(expr, fs)
        return simplify(res) if convert_everything else res

    ap = getattr(expr, "applyfunc", None)
    if callable(ap):
        return ap(
            lambda e: _convert_expr(e, conv, convert_everything=convert_everything)
        )

    return expr


# -----------------------------------------------------------------------------
# for tensors
# -----------------------------------------------------------------------------
def _coeff_dict_formatter(
    varSpace, coeff_dict, valence, total_degree, _varSpace_type, data_shape
):
    """
    Helper function to populate conversion dicts for tensor field classes
    """
    variable_registry = get_variable_registry()
    CVS = variable_registry["complex_variable_systems"]

    exhaust1 = list(varSpace)
    populate = {
        "compCoeffDataDict": dict(),
        "realCoeffDataDict": dict(),
        "holVarDict": dict(),
        "antiholVarDict": dict(),
        "realVarDict": dict(),
        "imVarDict": dict(),
        "preProcessMinDataToHol": dict(),
        "preProcessMinDataToReal": dict(),
    }
    if _varSpace_type == "real":
        for var in varSpace:
            varStr = str(var)
            if var in exhaust1:
                for parent in CVS.values():
                    if varStr in parent["variable_relatives"]:
                        cousin = (
                            set(
                                parent["variable_relatives"][varStr]["complex_family"][
                                    2:
                                ]
                            )
                            - {var}
                        ).pop()
                        if cousin in exhaust1:
                            exhaust1.remove(cousin)
                        if (
                            parent["variable_relatives"][varStr]["complex_positioning"]
                            == "real"
                        ):
                            realVar = var
                            exhaust1.remove(var)
                            imVar = cousin
                        else:
                            realVar = cousin
                            exhaust1.remove(var)
                            imVar = var
                        holVar = parent["variable_relatives"][varStr]["complex_family"][
                            0
                        ]
                        antiholVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][1]
                        populate["holVarDict"][holVar] = [realVar, imVar]
                        populate["antiholVarDict"][antiholVar] = [
                            realVar,
                            imVar,
                        ]
                        populate["realVarDict"][realVar] = [holVar, antiholVar]
                        populate["imVarDict"][imVar] = [holVar, antiholVar]
    else:  # _varSpace_type == 'complex'
        for var in varSpace:
            varStr = str(var)
            if var in exhaust1:
                for parent in CVS.values():
                    if varStr in parent["variable_relatives"]:
                        cousin = (
                            set(
                                parent["variable_relatives"][varStr]["complex_family"][
                                    :2
                                ]
                            )
                            - {var}
                        ).pop()
                        if cousin in exhaust1:
                            exhaust1.remove(cousin)
                        if (
                            parent["variable_relatives"][varStr]["complex_positioning"]
                            == "holomorphic"
                        ):
                            holVar = var
                            exhaust1.remove(var)
                            antiholVar = cousin
                        else:
                            holVar = cousin
                            exhaust1.remove(var)
                            antiholVar = var
                        realVar = parent["variable_relatives"][varStr][
                            "complex_family"
                        ][2]
                        imVar = parent["variable_relatives"][varStr]["complex_family"][
                            3
                        ]
                        populate["holVarDict"][holVar] = [realVar, imVar]
                        populate["antiholVarDict"][antiholVar] = [
                            realVar,
                            imVar,
                        ]
                        populate["realVarDict"][realVar] = [holVar, antiholVar]
                        populate["imVarDict"][imVar] = [holVar, antiholVar]
    new_realVarSpace = tuple(populate["realVarDict"].keys())
    new_holVarSpace = tuple(populate["holVarDict"].keys())
    new_antiholVarSpace = tuple(populate["antiholVarDict"].keys())
    new_imVarSpace = tuple(populate["imVarDict"].keys())

    if len(valence) == 0:
        if _varSpace_type == "real":
            populate["realCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["compCoeffDataDict"] = [
                new_holVarSpace + new_antiholVarSpace,
                {(0,) * total_degree: coeff_dict[(0,) * total_degree]},
            ]
        else:
            populate["compCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["realCoeffDataDict"] = [
                new_realVarSpace + new_imVarSpace,
                {(0,) * total_degree: coeff_dict[(0,) * total_degree]},
            ]
    else:

        def _retrieve_indices(term, typeSet=None):
            if typeSet == "symb":
                dictLoc = populate["realVarDict"] | populate["imVarDict"]
                refTuple = new_holVarSpace + new_antiholVarSpace
                termList = dictLoc[term]
            elif typeSet == "real":
                dictLoc = populate["holVarDict"] | populate["antiholVarDict"]
                refTuple = new_realVarSpace + new_imVarSpace
                termList = dictLoc[term]
            index_a = refTuple.index(termList[0])
            index_b = refTuple.index(termList[1], index_a + 1)
            return [index_a, index_b]

        # set up the conversion dicts for index conversion
        if _varSpace_type == "real":
            populate["preProcessMinDataToHol"] = {
                j: _retrieve_indices(varSpace[j], "symb") for j in range(len(varSpace))
            }

        else:  # if _varSpace_type == 'complex'
            populate["preProcessMinDataToReal"] = {
                j: _retrieve_indices(varSpace[j], "real") for j in range(len(varSpace))
            }

        # coordinate VF and DF conversion
        def decorateWithWeights(index, variance_rule, target="symb"):
            if variance_rule == 0:  # covariant case
                covariance = True
            else:  # contravariant case
                covariance = False

            if target == "symb":
                if (
                    varSpace[index]
                    in variable_registry["conversion_dictionaries"][
                        "real_part"
                    ].values()
                ):
                    holScale = (
                        exact_reciprocal(2) if covariance else 1
                    )  # D_z (d_z) coeff of D_x (d_x)
                    antiholScale = (
                        exact_reciprocal(2) if covariance else 1
                    )  # D_BARz (d_BARz) coeff of D_x (d_x)
                else:
                    holScale = (
                        -imag_unit() / 2 if covariance else imag_unit()
                    )  # D_z (d_z) coeff of D_y (d_y)
                    antiholScale = (
                        imag_unit() / 2 if covariance else -imag_unit()
                    )  # d_BARz (D_BARz) coeff of d_y (D_y)
                return [
                    [populate["preProcessMinDataToHol"][index][0], holScale],
                    [
                        populate["preProcessMinDataToHol"][index][1],
                        antiholScale,
                    ],
                ]
            else:  # converting from hol to real
                if (
                    varSpace[index]
                    in variable_registry["conversion_dictionaries"]["holToReal"]
                ):
                    realScale = (
                        1 if covariance else exact_reciprocal(2)
                    )  # D_x (d_x) coeff in D_z (d_z)
                    imScale = (
                        imag_unit()
                        if covariance
                        else -imag_unit() * exact_reciprocal(2)
                    )  # D_y (d_y) coeff in D_z (d_z)
                else:
                    realScale = (
                        1 if covariance else exact_reciprocal(2)
                    )  # D_x (d_x) coeff of D_BARz (d_BARz)
                    imScale = (
                        -imag_unit()
                        if covariance
                        else imag_unit() * exact_reciprocal(2)
                    )  # D_y (d_y) coeff of D_BARz (d_BARz)
                return [
                    [populate["preProcessMinDataToReal"][index][0], realScale],
                    [populate["preProcessMinDataToReal"][index][1], imScale],
                ]

        otherDict = dict()
        for term_index, term_coeff in coeff_dict.items():
            if _varSpace_type == "real":
                reformatTarget = "symb"
            else:
                reformatTarget = "real"
            termIndices = [
                decorateWithWeights(k, valence[j], target=reformatTarget)
                for j, k in enumerate(term_index)
            ]
            prodWithWeights = carProd_with_weights_without_R(*termIndices)
            prodWWRescaled = [[tuple(k[0]), term_coeff * k[1]] for k in prodWithWeights]
            minimal_term_set = _shape_basis(prodWWRescaled, data_shape)
            for term in minimal_term_set:
                if term[0] in otherDict:
                    oldVal = otherDict[term[0]]
                    otherDict[term[0]] = allToSym(oldVal + term[1])
                else:
                    otherDict[term[0]] = allToSym(term[1])

        if _varSpace_type == "real":
            populate["realCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["compCoeffDataDict"] = [
                new_holVarSpace + new_antiholVarSpace,
                otherDict,
            ]
        else:
            populate["compCoeffDataDict"] = [
                varSpace,
                coeff_dict,
            ]
            populate["realCoeffDataDict"] = [
                new_realVarSpace + new_imVarSpace,
                otherDict,
            ]

    return (
        populate,
        new_realVarSpace,
        new_holVarSpace,
        new_antiholVarSpace,
        new_imVarSpace,
    )


def _shape_basis(basis, shape):
    if shape == "symmetric":
        old_basis = dict(basis)
        new_basis = dict()
        for index, value in old_basis.items():
            new_index = tuple(sorted(index))
            if new_index in new_basis:
                new_basis[new_index] += value
            else:
                new_basis[new_index] = value
        return list(new_basis.items())
    if shape == "skew":
        old_basis = dict(basis)
        new_basis = dict()
        for index, value in old_basis.items():
            permS, new_index = permSign(index, returnSorted=True)
            new_index = tuple(new_index)
            if new_index in new_basis:
                new_basis[new_index] += permS * value
            else:
                new_basis[new_index] = permS * value
        return list(new_basis.items())
    return basis



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

from ._engine import _get_sage_module, _get_sympy_module, engine_kind
from ._symbolic_router import (
    _scalar_is_zero,
    as_numer_denom,
    cancel,
    expand,
    get_free_symbols,
)
from ._types_and_constants import symbol

__all__ = [
    "make_poly",
    "poly_gens",
    "poly_monoms",
    "poly_coeffs",
    "poly_total_degree",
    "is_polynomial",
    "extract_polynomial_coeffs",
    "poly_terms",
    "poly_linear_roots_from_factorization",
]


class PolyBackendError(RuntimeError):
    pass


def _stable_dedupe(seq: Iterable[Any]) -> Tuple[Any, ...]:
    out: List[Any] = []
    seen: set = set()
    for x in seq:
        if x in seen:
            continue
        seen.add(x)
        out.append(x)
    return tuple(out)


def _stable_dedupe_by_str(seq: Iterable[Any]) -> Tuple[Any, ...]:
    out: List[Any] = []
    seen: set = set()
    for x in seq:
        k = str(x)
        if k in seen:
            continue
        seen.add(k)
        out.append(x)
    return tuple(out)


def _unwrap_poly_expr(x: Any) -> Any:
    u = getattr(x, "poly_obj_unformatted", None)
    return u if u is not None else x


def _normalize_poly_expr(expr: Any, *, assume_polynomial: bool = False) -> Any:
    raw = _unwrap_poly_expr(expr)
    if not assume_polynomial:
        raw = cancel(raw)
        raw = as_numer_denom(raw)[0]
    return expand(raw)


@dataclass(frozen=True)
class _DGCVPolyTerms:
    expr: Any
    gens: Tuple[Any, ...]
    monoms: Tuple[Tuple[int, ...], ...]
    coeffs: Tuple[Any, ...]


def poly_terms(
    expr: Any,
    gens: Sequence[Any],
    *,
    assume_polynomial: bool = False,
    parameters: Optional[Iterable[Any]] = None,
) -> Tuple[Tuple[Any, ...], List[Tuple[int, ...]], List[Any]]:
    raw = _normalize_poly_expr(expr, assume_polynomial=assume_polynomial)
    kind = engine_kind()

    if kind == "sympy":
        gens_t = _stable_dedupe(gens)
        sp = _get_sympy_module()
        try:
            P = sp.Poly(raw, *gens_t)
        except Exception as e:
            raise PolyBackendError(
                f"dgcv: SymPy Poly construction failed for gens={gens_t!r} and expr type {type(raw).__name__}"
            ) from e
        monoms = [tuple(m) for m in P.monoms()]
        coeffs = list(P.coeffs())
        return gens_t, monoms, coeffs

    if kind == "sage":
        gens_t = _stable_dedupe_by_str(gens)
        monoms, coeffs = _sage_terms_via_coefficient(raw, gens_t)
        return gens_t, monoms, coeffs

    raise PolyBackendError(
        "dgcv: no supported symbolic engine is available for polynomials"
    )


def _sage_terms_via_coefficient(
    expr_sr: Any, gens: Tuple[Any, ...]
) -> Tuple[List[Tuple[int, ...]], List[Any]]:
    sage = _get_sage_module()
    SR = getattr(sage, "SR", None)
    if SR is None:
        raise PolyBackendError("dgcv: Sage backend active but sage.SR is unavailable")

    try:
        f = SR(expr_sr)
    except Exception:
        f = expr_sr

    def _coeff_dict_in_var(f0: Any, v: Any) -> Dict[int, Any]:
        deg = getattr(f0, "degree", None)
        if callable(deg):
            try:
                dmax = int(deg(v))
            except Exception:
                dmax = None
        else:
            dmax = None

        out: Dict[int, Any] = {}

        if dmax is not None and dmax >= 0:
            for k in range(dmax + 1):
                try:
                    ck = f0.coefficient(v, k)
                except Exception:
                    ck = None
                if ck is None or _scalar_is_zero(ck):
                    continue
                out[int(k)] = ck
            return out if out else {0: 0}

        try:
            _ = f0.coefficient(v, 0)
            _ = f0.coefficient(v, 1)
        except Exception:
            return {0: f0}

        cap = 256
        for k in range(cap + 1):
            try:
                ck = f0.coefficient(v, k)
            except Exception:
                break
            if ck is None:
                break
            if not _scalar_is_zero(ck):
                out[int(k)] = ck

        if not out:
            out[0] = f0
        return out

    def _rec(i: int, f0: Any) -> Dict[Tuple[int, ...], Any]:
        if i == len(gens):
            return {(): f0}
        v = gens[i]
        d = _coeff_dict_in_var(f0, v)
        out: Dict[Tuple[int, ...], Any] = {}
        for k, ck in d.items():
            tail = _rec(i + 1, ck)
            for e_tail, c_tail in tail.items():
                out[(int(k),) + e_tail] = c_tail
        return out

    term_dict = _rec(0, f)
    items = [(m, c) for (m, c) in term_dict.items() if not _scalar_is_zero(c)]
    items.sort(key=lambda mc: (sum(mc[0]), mc[0]))
    monoms = [tuple(int(e) for e in m) for m, _ in items]
    coeffs = [c for _, c in items]
    return monoms, coeffs


def extract_polynomial_coeffs(
    expr: Union[Any, Sequence[Any]],
    gens: Sequence[Any],
    *,
    assume_polynomial: bool = False,
    parameters: Optional[Iterable[Any]] = None,
    flatten: bool = True,
) -> List[Any]:
    exprs = list(expr) if isinstance(expr, (list, tuple)) else [expr]

    if flatten:
        out: List[Any] = []
        for e in exprs:
            _, _, coeffs = poly_terms(
                e, gens, assume_polynomial=assume_polynomial, parameters=parameters
            )
            out.extend(list(coeffs))
        return out

    grouped: List[List[Any]] = []
    for e in exprs:
        _, _, coeffs = poly_terms(
            e, gens, assume_polynomial=assume_polynomial, parameters=parameters
        )
        grouped.append(list(coeffs))
    return grouped  # type: ignore[return-value]


def make_poly(
    expr: Any, gens: Sequence[Any], *, parameters: Optional[Iterable[Any]] = None
) -> Any:
    kind = engine_kind()

    if kind == "sympy":
        gens_t = _stable_dedupe(gens)
        if len(gens_t) == 0 and len(get_free_symbols(expr)) == 0:
            gens_t = (symbol("_dgcv_atom"),)
        sp = _get_sympy_module()
        try:
            return sp.Poly(expr, *gens_t)
        except Exception as e:
            raise PolyBackendError(
                f"dgcv: SymPy Poly construction failed for gens={gens_t!r} and expr type {type(expr).__name__}"
            ) from e

    if kind == "sage":
        gens_t = _stable_dedupe_by_str(gens)
        raw = _normalize_poly_expr(expr, assume_polynomial=False)
        _gens_out, monoms, coeffs = poly_terms(
            raw, gens_t, assume_polynomial=True, parameters=parameters
        )
        return _DGCVPolyTerms(
            expr=raw,
            gens=tuple(_gens_out),
            monoms=tuple(tuple(int(e) for e in m) for m in monoms),
            coeffs=tuple(coeffs),
        )

    raise PolyBackendError(
        "dgcv: no supported symbolic engine is available for polynomials"
    )


def poly_gens(P: Any) -> Tuple[Any, ...]:
    kind = engine_kind()

    if kind == "sympy":
        try:
            return tuple(P.gens)
        except Exception as e:
            raise PolyBackendError("dgcv: failed to read SymPy Poly generators") from e

    if kind == "sage":
        if isinstance(P, _DGCVPolyTerms):
            return tuple(P.gens)
        try:
            return tuple(P.parent().gens())
        except Exception as e:
            raise PolyBackendError(
                "dgcv: failed to read Sage polynomial generators"
            ) from e

    raise PolyBackendError(
        "dgcv: no supported symbolic engine is available for polynomials"
    )


def poly_monoms(P: Any) -> List[Tuple[int, ...]]:
    kind = engine_kind()

    if kind == "sympy":
        try:
            return [tuple(m) for m in P.monoms()]
        except Exception as e:
            raise PolyBackendError(
                "dgcv: failed to extract SymPy polynomial monomials"
            ) from e

    if kind == "sage":
        if isinstance(P, _DGCVPolyTerms):
            return [tuple(int(e) for e in m) for m in P.monoms]
        try:
            d = P.dict()
            items = list(d.items())
            items.sort(key=lambda kv: (sum(kv[0]), kv[0]))
            return [tuple(int(i) for i in k) for k, _ in items]
        except Exception as e:
            raise PolyBackendError(
                "dgcv: failed to extract Sage polynomial monomials"
            ) from e

    raise PolyBackendError(
        "dgcv: no supported symbolic engine is available for polynomials"
    )


def poly_coeffs(P: Any) -> List[Any]:
    kind = engine_kind()

    if kind == "sympy":
        try:
            return list(P.coeffs())
        except Exception as e:
            raise PolyBackendError(
                "dgcv: failed to extract SymPy polynomial coefficients"
            ) from e

    if kind == "sage":
        if isinstance(P, _DGCVPolyTerms):
            return list(P.coeffs)
        try:
            d = P.dict()
            items = list(d.items())
            items.sort(key=lambda kv: (sum(kv[0]), kv[0]))
            return [v for _, v in items]
        except Exception as e:
            raise PolyBackendError(
                "dgcv: failed to extract Sage polynomial coefficients"
            ) from e

    raise PolyBackendError(
        "dgcv: no supported symbolic engine is available for polynomials"
    )


def poly_total_degree(
    expr: Any, gens: Sequence[Any], *, parameters: Optional[Iterable[Any]] = None
) -> Optional[int]:
    kind = engine_kind()

    if kind == "sympy":
        sp = _get_sympy_module()
        gens_t = _stable_dedupe(gens)
        try:
            return int(sp.total_degree(expr, *gens_t))
        except Exception:
            try:
                return int(sp.Poly(expr, *gens_t).total_degree())
            except Exception:
                return None

    if kind == "sage":
        try:
            P = make_poly(expr, gens, parameters=parameters)
            monoms = poly_monoms(P)
            if not monoms:
                return 0
            return max(sum(int(e) for e in m) for m in monoms)
        except Exception:
            return None

    return None


def is_polynomial(
    expr: Any, gens: Sequence[Any], *, parameters: Optional[Iterable[Any]] = None
) -> bool:
    try:
        make_poly(expr, gens, parameters=parameters)
        return True
    except Exception:
        return False


def poly_linear_roots_from_factorization(factored_poly: Any, var: Any) -> List[Any]:
    roots: List[Any] = []

    def _ordered_factors(expr: Any) -> List[Any]:
        aof = getattr(expr, "as_ordered_factors", None)
        if callable(aof):
            try:
                out = list(aof())
                return out if out else [expr]
            except Exception:
                pass

        try:
            if getattr(expr, "is_Mul", False):
                return list(getattr(expr, "args", ())) or [expr]
        except Exception:
            pass

        return [expr]

    def _unwrap_pow(f: Any) -> Any:
        try:
            if getattr(f, "is_Pow", False):
                base = getattr(f, "base", None)
                exp = getattr(f, "exp", None)
                if base is not None and exp is not None:
                    return base
        except Exception:
            pass

        try:
            if hasattr(f, "args") and len(f.args) == 2:
                base, _exp = f.args
                if str(type(f)).lower().find("pow") != -1:
                    return base
        except Exception:
            pass

        return f

    def _try_linear_root_from_poly(f: Any) -> Any:
        as_poly = getattr(f, "as_poly", None)
        if callable(as_poly):
            try:
                p = as_poly(var)
                if p is None:
                    return None
                deg = getattr(p, "degree", None)
                d = deg() if callable(deg) else None
                if d is None:
                    try:
                        d = p.degree()
                    except Exception:
                        d = None
                if d != 1:
                    return None
                all_coeffs = getattr(p, "all_coeffs", None)
                if callable(all_coeffs):
                    a, b = all_coeffs()
                else:
                    a, b = p.all_coeffs()
                if not _scalar_is_zero(a):
                    return -b / a
            except Exception:
                return None
        return None

    def _is_add_like(f: Any) -> bool:
        try:
            return bool(getattr(f, "is_Add", False))
        except Exception:
            return False

    def _try_linear_root_heuristic(f: Any) -> Any:
        try:
            args = getattr(f, "args", None)
            if args is not None and len(args) == 2:
                a0, a1 = args
                if a0 == var:
                    return -a1
                if a1 == var:
                    return -a0
        except Exception:
            pass

        if _is_add_like(f) or hasattr(f, "args"):
            try:
                terms = list(getattr(f, "args", ()))
                if not terms:
                    return None
            except Exception:
                return None

            var_term = None
            const_terms = []

            for t in terms:
                if t == var:
                    if var_term is None:
                        var_term = 1
                    else:
                        return None
                    continue

                try:
                    if getattr(t, "is_Mul", False):
                        mul_args = list(getattr(t, "args", ()))
                        if var in mul_args and mul_args.count(var) == 1:
                            coeff = 1
                            for u in mul_args:
                                if u == var:
                                    continue
                                coeff = coeff * u
                            if var_term is None:
                                var_term = coeff
                            else:
                                return None
                            continue
                except Exception:
                    pass

                const_terms.append(t)

            if var_term is None:
                return None

            b = 0 if not const_terms else const_terms[0]
            for u in const_terms[1:]:
                b = b + u

            try:
                if not _scalar_is_zero(var_term):
                    return -b / var_term
            except Exception:
                return None

        return None

    for f0 in _ordered_factors(factored_poly):
        f = _unwrap_pow(f0)

        r = _try_linear_root_from_poly(f)
        if r is None:
            r = _try_linear_root_heuristic(f)

        if r is not None:
            roots.append(r)

    out: List[Any] = []
    seen: set = set()
    for r in roots:
        k = str(r)
        if k in seen:
            continue
        seen.add(k)
        out.append(r)

    return out



"""
package: dgcv - Differential Geometry with Complex Variables
module: algebras/algebras_aux

Author (of this module): David Sykes (https://realandimaginary.com/dgcv/)

License:
    MIT License
"""

# -----------------------------------------------------------------------------
# imports
# -----------------------------------------------------------------------------
import numbers
import random

from .._config import dgcv_exception_note, dgcv_warning
from .._safeguards import (
    create_key,
    get_dgcv_category,
    get_dgcv_settings_registry,
    query_dgcv_categories,
    retrieve_passkey,
    retrieve_public_key,
)
from ..arrays import matrix_dgcv
from ..backends._numeric_router import _extract_basis_over_number_field
from ..backends._symbolic_router import (
    _scalar_is_zero,
    get_free_symbols,
    simplify,
    subs,
)
from ..backends._types_and_constants import is_atomic, rational, symbol
from ..dgcv_core import VF_bracket, allToReal, variableProcedure
from ..solvers import solve_dgcv
from ..vmf import clearVar, listVar


# -----------------------------------------------------------------------------
# utilities
# -----------------------------------------------------------------------------
def _as_matrix(obj):
    if isinstance(obj, matrix_dgcv):
        return obj
    try:
        return matrix_dgcv(obj)
    except Exception:
        return None


def _validate_structure_data(
    data,
    process_matrix_rep=False,
    assume_skew=False,
    assume_Lie_alg=False,
    basis_order_for_supplied_str_eqns=None,
    process_tensor_rep=False,
):
    if process_tensor_rep:
        # # DEBUG
        # return algebraDataFromTensorRep(data), "tensor"
        try:
            return algebraDataFromTensorRep(data), "tensor"
        except Exception as e:
            raise dgcv_exception_note(f"{e}") from None
    if process_matrix_rep:
        mats = []
        for obj in data:
            m = _as_matrix(obj)
            if m is None or m.nrows != m.ncols:
                mats = None
                break
            mats.append(m)
        if mats is not None:
            # # DEBUG:
            # return algebraDataFromMatRep(data), "matrix"
            try:
                return algebraDataFromMatRep(data), "matrix"
            except Exception as e:
                raise dgcv_exception_note(f"{e}") from None
        elif all(get_dgcv_category(elem) == "tensorProduct" for elem in data):
            dgcv_warning(
                "`_validate_structure_data` was given a list of tensorProduct instance, but `process_matrix_rep` was also marked True. The latter was ignored."
            )
            return _validate_structure_data(
                data,
                process_matrix_rep=False,
                assume_skew=assume_skew,
                assume_Lie_alg=assume_Lie_alg,
                basis_order_for_supplied_str_eqns=basis_order_for_supplied_str_eqns,
                process_tensor_rep=True,
            ), "tensor"
        else:
            raise ValueError(
                f"matrix representation processing requires a list of square matrices. Recieved: {data}"
            )

    if isinstance(data, (list, tuple)):
        if len(data) > 0:
            if all(query_dgcv_categories(obj, {"vector_field"}) for obj in data):
                return algebraDataFromVF(data)
        else:
            return tuple(), set()
    try:
        if isinstance(data, dict):
            if all(
                isinstance(key, tuple)
                and len(key) == 2
                and all(is_atomic(idx) for idx in key)
                for key in data
            ):
                if basis_order_for_supplied_str_eqns is None:
                    build_basis_order = True
                    basis_order_for_supplied_str_eqns = []
                else:
                    build_basis_order = False
                if not isinstance(
                    basis_order_for_supplied_str_eqns, (list, tuple)
                ) or not all(
                    is_atomic(var) for var in basis_order_for_supplied_str_eqns
                ):
                    raise ValueError(
                        "If initializing an algebra from structure equations and supplying the `basis_order_for_supplied_str_eqns` parameter, this parameter should be a list of the atomic variables appearing in the supplied structure equations."
                    )
                for var in set(sum([list(key) for key in data.keys()], [])):
                    if var not in basis_order_for_supplied_str_eqns:
                        if build_basis_order:
                            basis_order_for_supplied_str_eqns.append(var)
                        else:
                            raise ValueError(
                                "If initializing an algebra from structure equations and supplying the `basis_order_for_supplied_str_eqns` parameter, this parameter should be a list containing all atomic variables appearing in the supplied structure equations."
                            )
                ordered_BV = basis_order_for_supplied_str_eqns
                zeroing = {var: 0 for var in ordered_BV}
                new_data = dict()
                for idx_pair, val in data.items():
                    if not _scalar_is_zero(val):
                        v1, v2 = idx_pair
                        idx1 = ordered_BV.index(v1)
                        idx2 = ordered_BV.index(v2)
                        if hasattr(val, "subs") and _scalar_is_zero(val.subs(zeroing)):
                            coeffs = []
                            for var in ordered_BV:
                                coeffs.append(
                                    simplify(val.subs({var: 1}).subs(zeroing))
                                )
                            new_data[(idx2, idx1)] = tuple(coeffs)
                        else:
                            raise ValueError(
                                "If initializing an algebra from structure equations, supplied structure equations should be a dictionary whose keys are tuples of atomic variables and whose values are linear combinations of variables representing the product of the elements in the key tuple. If that is the case then you are likely getting this error because you did not supply the algebra creator with a valid value for the `basis_order_for_supplied_str_eqns` parameter. If that paremeter were omited, it is not always possible to unambiguously infer its proper value from general structure equations data, and hence this error arises."
                            )
                data = new_data
            if all(
                isinstance(key, tuple)
                and len(key) == 2
                and all(isinstance(idx, numbers.Integral) and idx >= 0 for idx in key)
                for key in data
            ):
                provided_index_bound = max(sum([list(key) for key in data.keys()], []))
            else:
                raise ValueError(
                    "Structure data must be have one of several formats: It can be a list/tuple with 3D shape of size (x, x, x). Or it can be a dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep."
                )
            if all(isinstance(val, (tuple, list)) for val in data.values()):
                base_dims = list(len(val) for val in data.values())
                if len(set(base_dims)) != 1 or base_dims[0] < provided_index_bound + 1:
                    raise ValueError(
                        "If initializing an algebra algebra with structure data from a dictionairy, its keys should be (i,j) index tuples and its values should be tuples of coefficients from the product of i and j basis elements. All values tuples must have the same length in particular. Indices in the keys must not exceed the length of value tuples - 1 (as indexing starts from 0!)"
                    )
                else:
                    base_dim = base_dims[0]
                if assume_skew or assume_Lie_alg:
                    seen = []
                    initial_keys = list(data.keys())
                    for idx in initial_keys:
                        if idx in seen:
                            pass
                        else:
                            invert_idx = (idx[1], idx[0])
                            if invert_idx in data.keys():
                                if any(
                                    j + k != 0
                                    for j, k in zip(data[idx], data[invert_idx])
                                ):
                                    raise ValueError(
                                        "Either `assume_skew=True` or `assume_Lie_alg=True` was passed to the algebra contructor, but the accompanying structure data was not skew symmetric."
                                    )
                            else:
                                data[invert_idx] = [-j for j in data[idx]]
                            seen += [idx, invert_idx]
                data = [
                    [list(data.get((j, k), [0] * base_dim)) for j in range(base_dim)]
                    for k in range(base_dim)
                ]
            else:
                raise ValueError(
                    "If initializing an algebra algebra with structure data from a dictionairy, its keys should be (i,j) index tuples and its values should be tuples of coefficients from the product of i and j basis elements. All values tuples must have the same length in particular."
                )
        params = set()

        def _tuple_scan(elems, par: set):
            for elem in elems:
                par |= getattr(elem, "free_symbols", set())
            return tuple(elems)

        # Check that the data is a 3D list-like structure
        if (
            isinstance(data, (list, tuple))
            and len(data) > 0
            and isinstance(data[0], (list, tuple))
        ):
            if len(data) == len(data[0]) == len(data[0][0]):
                sd = tuple(
                    tuple(_tuple_scan(inner, params) for inner in outer)
                    for outer in data
                )
                return sd, params
            else:
                raise ValueError(
                    "Structure data must be a list with 3D shape of size (x, x, x). Or it can a  dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep."
                )
        else:
            raise ValueError(
                "Structure data must be a list with 3D shape of size (x, x, x). Or it can a  dictionairy of the (i,j) entries for the structure data. Set `process_matrix_rep=True` to initialize from a matrix representation, or provide a list of vector fields to initialize from a VF rep."
            )
    except Exception as e:
        raise ValueError(f"Invalid structure data format: {type(data)} - {e}")


def algebraDataFromVF(
    vector_fields,
    *,
    samples=None,
    randomize_large=True,
    assume_basis=True,
    assume_consistent_coordinate_formatting=False,
    disable_sampling=False,
):
    vector_fields = list(vector_fields)
    dim0 = len(vector_fields)

    if dim0 == 0:
        return [], set()

    if not all(query_dgcv_categories(vf, {"vector_field"}) for vf in vector_fields):
        raise TypeError("algebraDataFromVF expects a list of vector fields.")

    vf_basis = vector_fields
    if not assume_basis:
        reduced = _extract_basis_over_number_field(vf_basis)
        if len(reduced) != len(vf_basis) and not get_dgcv_settings_registry().get(
            "forgo_warnings", False
        ):
            dgcv_warning(
                "algebraDataFromVF: vector field list was not linearly independent; "
                "reduced to an extracted basis. Use assume_basis=True to bypass."
            )
        vf_basis = reduced

    dim = len(vf_basis)
    if dim == 0:
        return [], set()

    if samples is None:
        samples = max(2, dim)

    atoms = set()
    for vf in vf_basis:
        vs = getattr(vf, "_variable_spaces", None)
        if isinstance(vs, dict):
            for tup in vs.values():
                atoms.update(tup)

    atoms = tuple(sorted(atoms, key=lambda x: str(x)))
    nvars = len(atoms)

    if randomize_large:

        def _rand_rat():
            return rational(random.randint(1, 1000), random.randint(1001, 2000))

    def _make_subs_dicts():
        if nvars == 0:
            return [None] * samples

        if randomize_large:
            return [
                {atoms[i]: _rand_rat() for i in range(nvars)} for _ in range(samples)
            ]

        return [
            {atoms[i]: rational((i + 1) ** (s + 1), 32) for i in range(nvars)}
            for s in range(samples)
        ]

    subs_dicts = _make_subs_dicts()
    if disable_sampling:
        subs_dicts = [None]

    def _scalar_eqns_from_objects(eqns):
        out = []
        for e in eqns:
            if get_dgcv_category(e) == "tensor_field":
                out.extend(e.coeff_dict.values())
            else:
                out.append(e)
        return out

    def _local_unknowns(prefix, n):
        label = create_key(prefix=prefix)
        return [symbol(f"{label}{i}") for i in range(n)]

    tvars = _local_unknowns("T", dim)

    combi = None
    for t, vf in zip(tvars, vf_basis):
        term = t * vf
        combi = term if combi is None else (combi + term)

    def _orig_index_pairs():
        if vf_basis is vector_fields:
            return list(range(dim)), list(range(dim))

        kept_indices = []
        start = 0
        for kept in vf_basis:
            found = False
            for i in range(start, len(vector_fields)):
                if vector_fields[i] is kept:
                    kept_indices.append(i)
                    start = i + 1
                    found = True
                    break
            if not found:
                kept_indices.append(None)
        return kept_indices, kept_indices

    kept_idx_map, _ = _orig_index_pairs()

    def compute_bracket(j, k):
        residual = VF_bracket(vf_basis[j], vf_basis[k]) - combi
        coeffs = list(residual.coeff_dict.values())

        eqs = []
        for subs_d in subs_dicts:
            for c in coeffs:
                v = c if subs_d is None else subs(c, subs_d)
                if not assume_consistent_coordinate_formatting:
                    v = allToReal(v)
                eqs.append(v)

        eqs = _scalar_eqns_from_objects(eqs)
        sols = solve_dgcv(eqs, tvars, method="linsolve")
        if not sols:
            jj = kept_idx_map[j] if j < len(kept_idx_map) else j
            kk = kept_idx_map[k] if k < len(kept_idx_map) else k
            raise RuntimeError(
                f"algebraDataFromVF: not closed under Lie bracket at indices ({jj}, {kk})."
            )

        sol = sols[0]
        out = []
        local_params = set()
        for t in tvars:
            coeff = subs(t, sol)
            fs = set(get_free_symbols(coeff))
            if atoms:
                fs -= set(atoms)
            local_params |= fs
            out.append(coeff)

        return out, local_params

    structure_data = [[[0 for _ in range(dim)] for _ in range(dim)] for _ in range(dim)]
    params = set()

    for j in range(dim):
        for k in range(j + 1, dim):
            c, p = compute_bracket(j, k)
            params |= p
            structure_data[j][k] = c
            structure_data[k][j] = [-elem for elem in c]

    return structure_data, params


def algebraDataFromMatRep(mat_list):
    """
    Create the structure data array for a Lie algebra from a list of matrices in *mat_list*.

    This function computes the Lie algebra structure constants from a matrix representation of a Lie algebra.
    The returned structure data can be used to initialize an algebra instance.

    Parameters
    ----------
    mat_list : list
        A list of square matrices of the same size representing the Lie algebra.

    Returns
    -------
    list
        A 3D list of lists of lists representing the Lie algebra structure data.

    Raises
    ------
    Exception
        If the matrices do not span a Lie algebra, or if the matrices are not square and of the same size.
    """
    if not isinstance(mat_list, (list, tuple)):
        raise Exception(
            "algorithm for extracting algebra data from matrices expects a list of square matrices."
        )

    mListLoc = []
    for j in mat_list:
        m = _as_matrix(j)
        if m is None:
            raise Exception(
                "algorithm for extracting algebra data from matrices expects a list of square matrices."
            )
        mListLoc.append(m)

    if not mListLoc:
        return ([], mat_list, set())

    shapeLoc = mListLoc[0].nrows
    indexRangeCap = len(mListLoc)

    if not all(m.shape == (shapeLoc, shapeLoc) for m in mListLoc):
        raise Exception(
            "algorithm for extracting algebra data from matrices expects a list of square matrices of the same size."
        )

    tempVarLabel = "T" + retrieve_public_key()
    vars = variableProcedure(
        tempVarLabel,
        indexRangeCap,
        return_created_object=True,
        _tempVar=retrieve_passkey(),
    )[0]

    combiMatLoc = matrix_dgcv.zeros(shapeLoc, shapeLoc)
    for j in range(indexRangeCap):
        combiMatLoc = combiMatLoc + vars[j] * mListLoc[j]

    params = set()

    def pairValue(j, k, par):
        mat = (mListLoc[j] @ mListLoc[k]) - (mListLoc[k] @ mListLoc[j]) - combiMatLoc

        bracketVals = list(set(mat._data.values()))
        if not bracketVals:
            return [0] * indexRangeCap
        if len(bracketVals) == 1 and _scalar_is_zero(bracketVals[0]):
            return [0] * indexRangeCap

        solLoc = list(solve_dgcv(bracketVals, vars))
        if len(solLoc) == 1:
            coeffs = []
            for var in vars:
                coeff = var.subs(solLoc[0])
                par |= get_free_symbols(coeff)
                coeffs.append(coeff)
            return coeffs

        clearVar(*listVar(temporary_only=True), report=False)
        raise Exception(
            f"Unable to determine if matrices are closed under commutators. "
            f"Problem matrices are in positions {j} and {k}."
        )

    structure_data = [
        [
            [0] * indexRangeCap if k <= j else pairValue(k, j, params)
            for j in range(indexRangeCap)
        ]
        for k in range(indexRangeCap)
    ]
    for k in range(indexRangeCap):
        for j in range(k + 1, indexRangeCap):
            structure_data[k][j] = [-entry for entry in structure_data[j][k]]

    clearVar(*listVar(temporary_only=True), report=False)

    return (structure_data, mat_list, params)


def algebraDataFromTensorRep(tensor_list):
    """
    Create the structure data array from a list of tensor products closed under the `_contraction_product` operator (see dgcv.tensorProduct documentation).

    Parameters
    ----------
    tensorProduct : list
        A list of tensorProduct instances

    Returns
    -------
    list
        A 3D array-like list of lists of lists representing the Lie algebra structure data.
    """

    tempVarLabel = "T" + create_key()
    dim = len(tensor_list)
    if dim == 0:
        return [[[]]], tensor_list, set()
    vars = variableProcedure(
        tempVarLabel, dim, return_created_object=True, _tempVar=retrieve_passkey()
    )[0]
    gen_elem = sum(
        [vars[j] * tensor_list[j] for j in range(1, dim)], vars[0] * tensor_list[0]
    )

    params = set()

    def computeBracket(j, k, par):
        if k < j:
            return [0] * dim
        product = (tensor_list[j] * tensor_list[k]) - gen_elem
        solutions = solve_dgcv(product, vars)
        if len(solutions) > 0:
            sol_values = solutions[0]
            coeffs = []
            for var in vars:
                coeff = var.subs(sol_values)
                par |= get_free_symbols(coeff)
                coeffs.append(coeff)
            return coeffs
        else:
            clearVar(*listVar(temporary_only=True), report=False)
            raise Exception(
                f"Contraction product of tensors at positions {j} and {k} are not in the given tensor list."
            )

    structure_data = [
        [[0 for _ in tensor_list] for _ in tensor_list] for _ in tensor_list
    ]

    for j in range(dim):
        for k in range(j):
            structure_data[k][j] = computeBracket(k, j, params)  # CHECK index order!!!
            structure_data[j][k] = [-elem for elem in structure_data[k][j]]

    clearVar(*listVar(temporary_only=True), report=False)

    return structure_data, tensor_list, params  # filter independants
